import copy
import functools
import io
from typing import Optional
from unittest.mock import MagicMock, Mock, patch

import numpy as np
import pytest
import transformers
from pandas.core.api import DataFrame as DataFrame
from PIL import Image
from typing_extensions import override

from oumi.builders import build_chat_template
from oumi.core.datasets.vision_language_dataset import VisionLanguageSftDataset
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)


class EqBytesIO:
    def __init__(self, bytes_io: io.BytesIO):
        self._byte_io = bytes_io

    def __eq__(self, other):
        return (
            isinstance(other, io.BytesIO)
            and other.getvalue() == self._byte_io.getvalue()
        )


_IMAGE_TOKEN = "<image_token>"
_IMAGE_TOKEN_ID = 32001


@pytest.fixture
def mock_image_tokenizer() -> MagicMock:
    def _convert_tokens_to_ids(token: str) -> int:
        if token == _IMAGE_TOKEN:
            return _IMAGE_TOKEN_ID
        return 101

    mock = MagicMock(spec=BaseTokenizer)
    mock.pad_token_id = 0
    mock.chat_template = build_chat_template("llava")
    mock.convert_tokens_to_ids = MagicMock(side_effect=_convert_tokens_to_ids)
    return mock


def create_mock_processor(label_ignore_index: Optional[int]):
    processor = Mock()
    processor.processor_name = "llava-hf/llava-1.5-7b-hf"
    processor.tokenizer = Mock()
    processor.image_processor = Mock()
    processor.chat_template = None
    processor.image_token = _IMAGE_TOKEN
    processor.image_token_id = _IMAGE_TOKEN_ID
    processor.label_ignore_index = label_ignore_index
    processor.side_effect = (
        lambda images, text, return_tensors, padding: transformers.BatchEncoding(
            data={
                "input_ids": [[101, 102, _IMAGE_TOKEN_ID, 104]],
                "attention_mask": [[1, 1, 1, 1]],
                "pixel_values": [
                    [
                        np.ones(shape=(3, 2, 8)),
                        np.zeros(shape=(3, 2, 8)),
                        np.ones(shape=(3, 2, 8)) * 0.5,
                        np.ones(shape=(3, 2, 8)) * 0.7,
                    ]
                ],
            },
            tensor_type=return_tensors,
        )
    )
    return processor


@pytest.fixture
def mock_processor():
    return create_mock_processor(label_ignore_index=-100)


@pytest.fixture
def mock_processor_no_label_ignore_index():
    return create_mock_processor(label_ignore_index=None)


@functools.cache  # same as @cache added in Python 3.9
def _get_test_png_image_bytes(image_size: Optional[tuple[int, int]] = None) -> bytes:
    if image_size is None:
        image_size = (80, 40)
    image = Image.new(mode="RGBA", size=image_size)
    bytes_io = io.BytesIO()
    image.save(bytes_io, format="PNG")
    return bytes_io.getvalue()


@pytest.fixture
def sample_conversation_using_image_path():
    return Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(content="Describe this image:", type=Type.TEXT),
                    ContentItem(content="path/to/image.jpg", type=Type.IMAGE_PATH),
                ],
            ),
            Message(
                role=Role.ASSISTANT,
                content="A beautiful sunset over the ocean.",
            ),
        ]
    )


@pytest.fixture
def sample_conversation_using_image_binary():
    return Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(content="Describe this image:", type=Type.TEXT),
                    ContentItem(
                        binary=_get_test_png_image_bytes(), type=Type.IMAGE_BINARY
                    ),
                ],
            ),
            Message(
                role=Role.ASSISTANT,
                content="A beautiful sunset over the ocean.",
            ),
        ]
    )


@pytest.fixture
def sample_dataset_image_path_no_label_ignore_index(
    mock_processor_no_label_ignore_index: Mock,
    sample_conversation_using_image_path: Conversation,
    mock_image_tokenizer: MagicMock,
):
    class TestDatasetImagePath(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            return sample_conversation_using_image_path

        @override
        def _load_data(self):
            pass

    return TestDatasetImagePath(
        processor=mock_processor_no_label_ignore_index,
        tokenizer=mock_image_tokenizer,
    )


@pytest.fixture
def sample_dataset_image_binary_label_ignore_index(
    mock_processor: Mock,
    sample_conversation_using_image_binary: Conversation,
    mock_image_tokenizer: MagicMock,
):
    class TestDatasetImageBinary(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            return sample_conversation_using_image_binary

        @override
        def _load_data(self):
            pass

    return TestDatasetImageBinary(
        processor=mock_processor, tokenizer=mock_image_tokenizer
    )


@pytest.fixture
def sample_dataset_image_binary_return_conversations_barebones(
    sample_conversation_using_image_binary: Conversation,
):
    class TestDatasetImageBinary(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            return sample_conversation_using_image_binary

        @override
        def _load_data(self):
            pass

    return TestDatasetImageBinary(
        # No processor/tokenizer
        return_conversations=True,
    )


@pytest.fixture
def sample_dataset_image_binary_return_conversations_with_processor(
    mock_processor: Mock,
    sample_conversation_using_image_binary: Conversation,
    mock_image_tokenizer: MagicMock,
):
    class TestDatasetImageBinary(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            return sample_conversation_using_image_binary

        @override
        def _load_data(self):
            pass

    return TestDatasetImageBinary(
        processor=mock_processor,
        tokenizer=mock_image_tokenizer,
        return_conversations=True,
    )


@pytest.fixture
def return_conversation_fixtures(
    sample_dataset_image_binary_return_conversations_barebones,
    sample_dataset_image_binary_return_conversations_with_processor,
):
    return {
        "barebones": (sample_dataset_image_binary_return_conversations_barebones),
        "with_processor": (
            sample_dataset_image_binary_return_conversations_with_processor
        ),
    }


def test_transform_simple_model_using_image_path(
    sample_dataset_image_path_no_label_ignore_index,
):
    my_dataset = sample_dataset_image_path_no_label_ignore_index
    with patch.object(
        my_dataset._feature_generator,
        "_load_image",
    ) as mock_load_image:
        mock_image = Mock(spec=Image.Image)
        mock_load_image.return_value = mock_image

        result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert "input_ids" in result
    assert np.array(result["input_ids"]).shape == (4,)
    assert "attention_mask" in result
    assert np.array(result["attention_mask"]).shape == (4,)
    assert "labels" in result
    assert np.array(result["labels"]).shape == (4,)
    assert np.all(np.array(result["labels"]) == np.array(result["input_ids"]))
    assert "pixel_values" in result
    assert np.array(result["pixel_values"]).shape == (4, 3, 2, 8)


def test_transform_simple_model_using_image_binary(
    sample_dataset_image_binary_label_ignore_index,
):
    my_dataset = sample_dataset_image_binary_label_ignore_index
    with patch.object(my_dataset._feature_generator, "_load_image") as mock_load_image:
        mock_image = Mock(spec=Image.Image)
        mock_load_image.return_value = mock_image

        result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert "input_ids" in result
    assert np.array(result["input_ids"]).shape == (4,)
    assert np.all(
        np.array(result["input_ids"]) == np.array([101, 102, _IMAGE_TOKEN_ID, 104])
    )
    assert "attention_mask" in result
    assert np.array(result["attention_mask"]).shape == (4,)
    assert "labels" in result
    assert np.array(result["labels"]).shape == (4,)
    assert np.all(np.array(result["labels"]) == np.array([101, 102, -100, 104]))
    assert "pixel_values" in result
    assert np.array(result["pixel_values"]).shape == (4, 3, 2, 8)


def test_transform_instruct_model_using_image_path(
    sample_dataset_image_path_no_label_ignore_index,
    mock_processor_no_label_ignore_index: Mock,
):
    my_dataset = sample_dataset_image_path_no_label_ignore_index
    mock_processor_no_label_ignore_index.chat_template = "Template"
    mock_processor_no_label_ignore_index.apply_chat_template = Mock(
        return_value="Processed template"
    )

    with patch.object(my_dataset._feature_generator, "_load_image") as mock_load_image:
        mock_image = Mock(spec=Image.Image)
        mock_load_image.return_value = mock_image

        result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert "input_ids" in result
    assert np.array(result["input_ids"]).shape == (4,)
    assert "attention_mask" in result
    assert np.array(result["attention_mask"]).shape == (4,)
    assert "labels" in result
    assert np.array(result["labels"]).shape == (4,)
    assert np.all(np.array(result["labels"]) == np.array(result["input_ids"]))
    assert "pixel_values" in result
    assert np.array(result["pixel_values"]).shape == (4, 3, 2, 8)
    mock_processor_no_label_ignore_index.apply_chat_template.assert_called_once()


def test_transform_instruct_model_using_image_binary(
    sample_dataset_image_binary_label_ignore_index, mock_processor: Mock
):
    my_dataset = sample_dataset_image_binary_label_ignore_index
    mock_processor.chat_template = "Template"
    mock_processor.apply_chat_template = Mock(return_value="Processed template")

    with patch.object(my_dataset._feature_generator, "_load_image") as mock_load_image:
        mock_image = Mock(spec=Image.Image)
        mock_load_image.return_value = mock_image

        result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert "input_ids" in result
    assert np.array(result["input_ids"]).shape == (4,)
    assert np.all(
        np.array(result["input_ids"]) == np.array([101, 102, _IMAGE_TOKEN_ID, 104])
    )
    assert "attention_mask" in result
    assert np.array(result["attention_mask"]).shape == (4,)
    assert "labels" in result
    assert np.array(result["labels"]).shape == (4,)
    assert np.all(np.array(result["labels"]) == np.array([101, 102, -100, 104]))
    assert "pixel_values" in result
    assert np.array(result["pixel_values"]).shape == (4, 3, 2, 8)
    mock_processor.apply_chat_template.assert_called_once()


@pytest.mark.parametrize(
    "fixture_name",
    ["barebones", "with_processor"],
)
def test_return_conversations(
    fixture_name: str,
    return_conversation_fixtures,
    mock_processor: Mock,
    sample_conversation_using_image_binary: Conversation,
):
    my_dataset = return_conversation_fixtures[fixture_name]
    mock_processor.chat_template = "Template"
    mock_processor.apply_chat_template = Mock(return_value="Processed template")

    assert my_dataset._feature_generator is None

    result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert set({"conversation_json"}) == set(result.keys())
    assert "conversation_json" in result
    conversation = Conversation.from_json(result["conversation_json"])
    assert conversation == sample_conversation_using_image_binary

    mock_processor.apply_chat_template.assert_not_called()


def test_return_conversations_with_max_images(
    mock_processor,
    mock_image_tokenizer,
    sample_conversation_using_image_binary: Conversation,
):
    class TestDatasetImageBinary(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            convo = sample_conversation_using_image_binary
            messages = copy.deepcopy(convo.messages)
            last_message = messages[-1]
            assert isinstance(last_message.content, str)
            messages[-1] = Message(
                role=last_message.role,
                content=[
                    ContentItem(content=last_message.content, type=Type.TEXT),
                    ContentItem(content="/anotherimage.jpg", type=Type.IMAGE_PATH),
                    ContentItem(
                        content="http://oumi.ai/yet_another_image.png",
                        type=Type.IMAGE_URL,
                    ),
                ],
            )
            return Conversation(
                conversation_id=convo.conversation_id,
                messages=messages,
                metadata=convo.metadata,
            )

        @override
        def _load_data(self):
            pass

    my_dataset = TestDatasetImageBinary(
        processor=mock_processor,
        tokenizer=mock_image_tokenizer,
        return_conversations=True,
        max_images=1,
    )

    mock_processor.chat_template = "Template"
    mock_processor.apply_chat_template = Mock(return_value="Processed template")

    assert my_dataset._feature_generator is None

    result = my_dataset.transform({"example": "data"})

    assert isinstance(result, dict)
    assert set({"conversation_json"}) == set(result.keys())
    assert "conversation_json" in result
    conversation = Conversation.from_json(result["conversation_json"])
    assert conversation == sample_conversation_using_image_binary

    mock_processor.apply_chat_template.assert_not_called()


def test_dataset_no_tokenizer(
    mock_processor: Mock,
    sample_conversation_using_image_binary: Conversation,
):
    class FooDataset(VisionLanguageSftDataset):
        default_dataset = "custom"

        @override
        def transform_conversation(self, example):
            return sample_conversation_using_image_binary

        @override
        def _load_data(self):
            pass

    with pytest.raises(ValueError, match="Tokenizer must be provided"):
        FooDataset(processor=mock_processor)

    with pytest.raises(ValueError, match="Tokenizer must be provided"):
        FooDataset(processor=mock_processor, tokenizer=None)



import os
import tempfile
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
import typer
import yaml
from requests.exceptions import RequestException
from typer.testing import CliRunner

from oumi.cli.cli_utils import (
    CONFIG_FLAGS,
    CONTEXT_ALLOW_EXTRA_ARGS,
    LogLevel,
    configure_common_env_vars,
    parse_extra_cli_args,
    resolve_and_fetch_config,
    section_header,
)


@pytest.fixture
def mock_response():
    response = Mock()
    response.text = "key: value"
    response.raise_for_status.return_value = None
    return response


@pytest.fixture
def mock_requests(mock_response):
    with patch("oumi.cli.cli_utils.requests") as r_mock:
        r_mock.get.return_value = mock_response
        yield r_mock


def simple_command(ctx: typer.Context):
    print(str(parse_extra_cli_args(ctx)))


runner = CliRunner()


#
# Fixtures
#
@pytest.fixture
def app():
    fake_app = typer.Typer()
    fake_app.command(context_settings=CONTEXT_ALLOW_EXTRA_ARGS)(simple_command)
    yield fake_app


def test_config_flags():
    # Simple test to ensure that this constant isn't changed accidentally.
    assert CONFIG_FLAGS == ["--config", "-c"]


def test_context_allow_extra_args():
    # Simple test to ensure that this constant isn't changed accidentally.
    assert CONTEXT_ALLOW_EXTRA_ARGS == {
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    }


def test_parse_extra_cli_args_space_separated(app):
    # Verify that results are in the proper dot format.
    result = runner.invoke(app, ["--config", "some/path", "--allow_extra", "args"])
    expected_result = ["config=some/path", "allow_extra=args"]
    assert result.output.strip() == str(expected_result).strip()


def test_parse_extra_cli_args_eq_separated(app):
    # Verify that results are in the proper dot format.
    result = runner.invoke(app, ["--config=some/path", "--allow_extra=args"])
    expected_result = ["config=some/path", "allow_extra=args"]
    assert result.output.strip() == str(expected_result).strip()


def test_parse_extra_cli_args_mixed(app):
    # Verify that results are in the proper dot format.
    result = runner.invoke(
        app, ["--config=some/path", "--foo ", " bar ", "--bazz = 12345 ", "--zz=XYZ"]
    )
    expected_result = ["config=some/path", "foo=bar", "bazz=12345", "zz=XYZ"]
    assert result.output.strip() == str(expected_result).strip()


def test_parse_extra_cli_args_empty(app):
    result = runner.invoke(app, [])
    expected_result = "[]"
    assert result.output.strip() == str(expected_result).strip()


def test_parse_extra_cli_args_fails_for_odd_args(app):
    # Verify that results are in the proper dot format.
    result = runner.invoke(app, ["--config", "some/path", "--odd"])
    output_str = result.output.strip()
    assert "Trailing argument has no value assigned" in output_str, f"{output_str}"


def test_valid_log_levels():
    # Verify that the log levels are valid.
    expected_levels = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
    supported_levels = set(LogLevel.__members__.keys())
    assert expected_levels == supported_levels


@patch.dict(os.environ, {"FOO": "1"}, clear=True)
def test_configure_common_env_vars_empty():
    configure_common_env_vars()
    assert os.environ == {
        "FOO": "1",
        "ACCELERATE_LOG_LEVEL": "info",
        "TOKENIZERS_PARALLELISM": "false",
    }


@patch.dict(
    os.environ,
    {
        "TOKENIZERS_PARALLELISM": "true",
        "FOO": "1",
    },
    clear=True,
)
def test_configure_common_env_vars_partially_preconfigured():
    configure_common_env_vars()
    assert os.environ == {
        "FOO": "1",
        "ACCELERATE_LOG_LEVEL": "info",
        "TOKENIZERS_PARALLELISM": "true",
    }


@patch.dict(
    os.environ,
    {"TOKENIZERS_PARALLELISM": "true", "ACCELERATE_LOG_LEVEL": "debug"},
    clear=True,
)
def test_configure_common_env_vars_fully_preconfigured():
    configure_common_env_vars()
    assert os.environ == {
        "ACCELERATE_LOG_LEVEL": "debug",
        "TOKENIZERS_PARALLELISM": "true",
    }


def test_resolve_and_fetch_config_with_oumi_prefix_and_explicit_output_dir(
    mock_requests,
):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"

        # When
        result = resolve_and_fetch_config(config_path, output_dir)

        # Then
        assert result == expected_path
        mock_requests.get.assert_called_once()
        assert expected_path.exists()


def test_resolve_and_fetch_config_without_prefix_and_explicit_output_dir(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = (
            "configs/recipes/smollm/inference/135m_infer.yaml"  # No oumi:// prefix
        )

        # When
        result = resolve_and_fetch_config(config_path, output_dir)

        # Then
        assert result == Path(config_path)
        assert not mock_requests.get.called


def test_resolve_and_fetch_config_with_oumi_prefix_and_env_dir(
    mock_requests, monkeypatch
):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = (
            Path(temp_dir) / "configs/recipes/smollm/inference/135m_infer.yaml"
        )
        monkeypatch.setenv("OUMI_DIR", temp_dir)

        # When
        result = resolve_and_fetch_config(config_path)

        # Then
        assert result == expected_path
        mock_requests.get.assert_called_once()
        assert expected_path.exists()


def test_resolve_and_fetch_config_without_prefix_and_env_dir(
    mock_requests, monkeypatch
):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        config_path = (
            "configs/recipes/smollm/inference/135m_infer.yaml"  # No oumi:// prefix
        )
        monkeypatch.setenv("OUMI_DIR", temp_dir)

        # When
        result = resolve_and_fetch_config(config_path)

        # Then
        assert result == Path(config_path)
        assert not mock_requests.get.called


def test_resolve_and_fetch_config_with_oumi_prefix_and_default_dir(
    mock_requests, monkeypatch
):
    with tempfile.TemporaryDirectory() as temp_dir:
        with patch("oumi.cli.cli_utils.OUMI_FETCH_DIR", temp_dir):
            # Given
            config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
            expected_path = (
                Path(temp_dir) / "configs/recipes/smollm/inference/135m_infer.yaml"
            )
            monkeypatch.delenv("OUMI_DIR", raising=False)

            # When
            result = resolve_and_fetch_config(config_path)

            # Then
            assert result == expected_path
            mock_requests.get.assert_called_once()
            assert expected_path.exists()


def test_resolve_and_fetch_config_without_prefix_and_default_dir(
    mock_requests, monkeypatch
):
    with tempfile.TemporaryDirectory() as temp_dir:
        with patch("oumi.cli.cli_utils.OUMI_FETCH_DIR", temp_dir):
            # Given
            config_path = (
                "configs/recipes/smollm/inference/135m_infer.yaml"  # No oumi:// prefix
            )
            expected_path = Path(config_path)
            monkeypatch.delenv("OUMI_DIR", raising=False)

            # When
            result = resolve_and_fetch_config(config_path)

            # Then
            assert result == expected_path
            assert not mock_requests.get.called


def test_resolve_and_fetch_config_with_existing_file_default_force(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"

        # Create existing file
        expected_path.parent.mkdir(parents=True)
        expected_path.write_text("existing content")

        # When
        result = resolve_and_fetch_config(config_path, output_dir)
        # Then
        assert result == expected_path
        assert mock_requests.get.call_count == 1
        assert expected_path.read_text() == "key: value"


def test_resolve_and_fetch_config_with_existing_file_force(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"

        # Create existing file
        expected_path.parent.mkdir(parents=True)
        expected_path.write_text("existing content")

        # When
        result = resolve_and_fetch_config(config_path, output_dir, force=True)

        # Then
        assert result == expected_path
        mock_requests.get.assert_called_once()
        assert expected_path.exists()
        assert expected_path.read_text() == "key: value"  # From mock_response


def test_resolve_and_fetch_config_force_no_conflict(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        expected_path = output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"

        # When
        result = resolve_and_fetch_config(config_path, output_dir, force=True)
        # Then
        assert result == expected_path
        assert mock_requests.get.call_count == 1
        assert expected_path.read_text() == "key: value"


def test_resolve_and_fetch_config_conflict_no_force(mock_requests):
    with pytest.raises(RuntimeError, match="Config already exists at"):
        with tempfile.TemporaryDirectory() as temp_dir:
            # Given
            output_dir = Path(temp_dir)
            config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
            expected_path = (
                output_dir / "configs/recipes/smollm/inference/135m_infer.yaml"
            )

            # Create existing file
            expected_path.parent.mkdir(parents=True)
            expected_path.write_text("existing content")

            # When
            result = resolve_and_fetch_config(config_path, output_dir, force=False)
            # Then
            assert result == expected_path
            assert mock_requests.get.call_count == 1
            assert expected_path.read_text() == "key: value"


def test_resolve_and_fetch_config_http_error(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        mock_requests.get.side_effect = RequestException("HTTP Error")

        # When
        with pytest.raises(RequestException):
            _ = resolve_and_fetch_config(config_path, output_dir, force=False)


def test_resolve_and_fetch_config_yaml_error(mock_requests):
    with tempfile.TemporaryDirectory() as temp_dir:
        # Given
        output_dir = Path(temp_dir)
        config_path = "oumi://configs/recipes/smollm/inference/135m_infer.yaml"
        mock_requests.get.return_value.text = "foo: bar\nbye"
        # When
        with pytest.raises(yaml.YAMLError):
            _ = resolve_and_fetch_config(config_path, output_dir, force=False)


def test_section_header():
    """Test the section_header function."""

    # Create a mock console
    mock_console = Mock()
    mock_console.width = 10

    # Call the function with a test message
    test_message = "Test Header"
    section_header(test_message, console=mock_console)

    # Assert that the console's print method was called with the correct message
    mock_console.print.assert_has_calls(
        [
            call("\n[blue]━━━━━━━━━━[/blue]"),
            call("[yellow]   Test Header[/yellow]"),
            call("[blue]━━━━━━━━━━[/blue]\n"),
        ]
    )



import json
import os
import tempfile

import pytest

from oumi.core.configs import EvaluationConfig, EvaluationTaskParams
from oumi.core.configs.inference_engine_type import InferenceEngineType
from oumi.core.configs.params.generation_params import GenerationParams
from oumi.core.configs.params.model_params import ModelParams
from oumi.core.configs.params.remote_params import RemoteParams
from oumi.core.evaluation.evaluation_result import EvaluationResult
from oumi.core.evaluation.utils.save_utils import (
    OUTPUT_FILENAME_BACKEND_CONFIG,
    OUTPUT_FILENAME_GENERATION_PARAMS,
    OUTPUT_FILENAME_INFERENCE_PARAMS,
    OUTPUT_FILENAME_MODEL_PARAMS,
    OUTPUT_FILENAME_PACKAGE_VERSIONS,
    OUTPUT_FILENAME_TASK_PARAMS,
    OUTPUT_FILENAME_TASK_RESULT,
    save_evaluation_output,
)


def test_save_evaluation_output_happy_path():
    # Set input parameters.
    task_params = EvaluationTaskParams(
        evaluation_backend="lm_harness",
        task_name="mmlu",
        num_samples=5,
        log_samples=True,
        eval_kwargs={"num_fewshot": 5},
    )
    result = EvaluationResult(
        task_name="NOT_USED",
        task_result={"my_metric": 10},
        backend_config={"config_param": 11},
        start_time="20250303_094257",
        elapsed_time_sec=15,
    )
    config = EvaluationConfig(
        tasks=[],
        model=ModelParams(model_name="my_model"),
        generation=GenerationParams(max_new_tokens=8),
        inference_engine=InferenceEngineType.VLLM,
        inference_remote_params=RemoteParams(api_url="my_api_url"),
    )

    with tempfile.TemporaryDirectory() as output_temp_dir:
        my_base_output_dir = os.path.join(output_temp_dir, "my_base_output_dir")

        # Run the function to be tested.
        save_evaluation_output(
            backend_name="mock_backend_name",
            task_params=task_params,
            evaluation_result=result,
            base_output_dir=my_base_output_dir,
            config=config,
        )

        # Ensure the output path was created.
        output_path = os.path.join(
            my_base_output_dir, f"mock_backend_name_{result.start_time}"
        )
        assert os.path.exists(output_path)

        # Validate the task result from the test's output file.
        task_result_path = os.path.join(output_path, OUTPUT_FILENAME_TASK_RESULT)
        with open(task_result_path, encoding="utf-8") as file_ptr:
            task_result_dict = json.load(file_ptr)
        assert task_result_dict == {
            "my_metric": 10,
            "start_time": "20250303_094257",
            "duration_sec": 15,
        }

        # Validate the backend config from the test's output file.
        backend_config_path = os.path.join(output_path, OUTPUT_FILENAME_BACKEND_CONFIG)
        with open(backend_config_path, encoding="utf-8") as file_ptr:
            backend_config_dict = json.load(file_ptr)
        assert backend_config_dict == {"config_param": 11}

        # Validate the task params from the test's output file.
        task_params_path = os.path.join(output_path, OUTPUT_FILENAME_TASK_PARAMS)
        with open(task_params_path, encoding="utf-8") as file_ptr:
            task_params_dict = json.load(file_ptr)
        assert task_params_dict == {
            "evaluation_backend": "lm_harness",
            "task_name": "mmlu",
            "num_samples": 5,
            "log_samples": True,
            "eval_kwargs": {"num_fewshot": 5},
        }

        # Validate the model params from the test's output file.
        model_params_path = os.path.join(output_path, OUTPUT_FILENAME_MODEL_PARAMS)
        with open(model_params_path, encoding="utf-8") as file_ptr:
            model_params_dict = json.load(file_ptr)
        assert "model_name" in model_params_dict
        assert model_params_dict["model_name"] == "my_model"

        # Validate the generation params from the test's output file.
        generation_params_path = os.path.join(
            output_path, OUTPUT_FILENAME_GENERATION_PARAMS
        )
        with open(generation_params_path, encoding="utf-8") as file_ptr:
            generation_params_dict = json.load(file_ptr)
        assert "max_new_tokens" in generation_params_dict
        assert generation_params_dict["max_new_tokens"] == 8

        # Validate the inference params from the test's output file.
        inference_params_path = os.path.join(
            output_path, OUTPUT_FILENAME_INFERENCE_PARAMS
        )
        with open(inference_params_path, encoding="utf-8") as file_ptr:
            inference_params_dict = json.load(file_ptr)
        assert "engine" in inference_params_dict
        assert inference_params_dict["engine"] == "VLLM"
        assert "remote_params" in inference_params_dict
        assert inference_params_dict["remote_params"]
        assert "api_url='my_api_url'" in inference_params_dict["remote_params"]

        # Validate the package versions from the test's output file.
        package_versions_path = os.path.join(
            output_path, OUTPUT_FILENAME_PACKAGE_VERSIONS
        )
        with open(package_versions_path, encoding="utf-8") as file_ptr:
            package_versions_dict = json.load(file_ptr)
        assert "oumi" in package_versions_dict


def test_save_evaluation_output_empty_results():
    with tempfile.TemporaryDirectory() as output_temp_dir:
        my_base_output_dir = os.path.join(output_temp_dir, "my_base_output_dir")

        # Run the function to be tested.
        save_evaluation_output(
            backend_name="mock_backend_name",
            task_params=EvaluationTaskParams(evaluation_backend="lm_harness"),
            evaluation_result=EvaluationResult(),
            base_output_dir=my_base_output_dir,
            config=EvaluationConfig(),
        )

        # Ensure the output directory was created.
        output_path = os.path.join(my_base_output_dir, "mock_backend_name")
        assert os.path.exists(output_path)
        output_files = list(os.listdir(output_path))

    # Ensure all parames have been saved.
    expected_files = [
        OUTPUT_FILENAME_TASK_PARAMS,
        OUTPUT_FILENAME_MODEL_PARAMS,
        OUTPUT_FILENAME_GENERATION_PARAMS,
        OUTPUT_FILENAME_INFERENCE_PARAMS,
        OUTPUT_FILENAME_PACKAGE_VERSIONS,
    ]
    for file in expected_files:
        assert file in output_files

    # Ensure that the task result and backend config files were NOT created.
    assert OUTPUT_FILENAME_TASK_RESULT not in output_files
    assert OUTPUT_FILENAME_BACKEND_CONFIG not in output_files


def test_save_evaluation_output_missing_backend():
    with pytest.raises(
        ValueError,
        match="The evaluation backend name must be provided.",
    ):
        save_evaluation_output(
            backend_name="",
            task_params=EvaluationTaskParams(evaluation_backend="lm_harness"),
            evaluation_result=EvaluationResult(),
            base_output_dir="",
            config=EvaluationConfig(),
        )


@pytest.mark.parametrize(
    "backend_dir, preexisting_dirs, expected_dir",
    [
        ("backend_dir", ["backend_dir"], "backend_dir_1"),
        ("backend_dir", ["backend_dir", "backend_dir_1"], "backend_dir_2"),
        (
            "backend_dir",
            ["backend_dir", "backend_dir_1", "backend_dir_2"],
            "backend_dir_3",
        ),
    ],
    ids=[
        "test_save_evaluation_output_with_a_preexisting_dir",
        "test_save_evaluation_output_with_2_preexisting_dirs",
        "test_save_evaluation_output_with_3_preexisting_dirs",
    ],
)
def test_save_evaluation_output_with_preexisting_path(
    backend_dir: str,
    preexisting_dirs: list[str],
    expected_dir: str,
) -> None:
    with tempfile.TemporaryDirectory() as output_temp_dir:
        # Create directories that the test requires to pre-exist.
        for preexisting_dir in preexisting_dirs:
            preexisting_path = os.path.join(output_temp_dir, preexisting_dir)
            os.makedirs(preexisting_path)

        # Run the save function to create a new dictionary.
        save_evaluation_output(
            backend_name=backend_dir,
            task_params=EvaluationTaskParams(evaluation_backend="lm_harness"),
            evaluation_result=EvaluationResult(),
            base_output_dir=output_temp_dir,
            config=EvaluationConfig(),
        )

        # Find the generated dictionary name and ensure it's correct.
        paths = list(os.listdir(output_temp_dir))
        assert len(paths) == len(preexisting_dirs) + 1
        new_dir = list(set(paths) - set(preexisting_dirs))
        assert len(new_dir) == 1
        new_dir = new_dir[0]
        assert new_dir == expected_dir



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override

from oumi.core.datasets import VisionLanguageSftDataset
from oumi.core.registry import register_dataset
from oumi.core.types.conversation import (
    ContentItem,
    Conversation,
    Message,
    Role,
    Type,
)


@register_dataset("nlphuji/flickr30k")
class Flickr30kDataset(VisionLanguageSftDataset):
    """Dataset class for the `nlphuji/flickr30k` dataset."""

    default_dataset = "nlphuji/flickr30k"

    @override
    def transform_conversation(self, example: dict) -> Conversation:
        """Transform a single conversation example into a Conversation object."""
        input_text = "Describe this image:"
        output_text = example["caption"][0]

        return Conversation(
            messages=[
                Message(
                    role=Role.USER,
                    content=[
                        ContentItem(
                            type=Type.IMAGE_BINARY,
                            binary=example["image"]["bytes"],
                        ),
                        ContentItem(type=Type.TEXT, content=input_text),
                    ],
                ),
                Message(role=Role.ASSISTANT, content=output_text),
            ]
        )



import base64
from typing import Final, cast

import pydantic
import pytest

from oumi.core.types.conversation import (
    ContentItem,
    ContentItemCounts,
    Conversation,
    Message,
    Role,
    Type,
)
from oumi.utils.image_utils import load_image_png_bytes_from_path

_SMALL_B64_IMAGE: Final[str] = (
    "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
)


def _create_test_image_bytes() -> bytes:
    return base64.b64decode(_SMALL_B64_IMAGE)


@pytest.fixture
def test_conversation():
    role_user = Role.USER
    role_assistant = Role.ASSISTANT

    message1 = Message(role=role_user, content="Hello", id="1")
    message2 = Message(role=role_assistant, content="Hi, how can I help you?")
    message3 = Message(
        role=role_user,
        content=[
            ContentItem(type=Type.TEXT, content="I need assistance with my account."),
            ContentItem(type=Type.IMAGE_BINARY, binary=_create_test_image_bytes()),
        ],
    )

    conversation = Conversation(messages=[message1, message2, message3])
    return conversation, role_user, role_assistant, message1, message2, message3


def test_first_message_no_role(test_conversation):
    conversation, _, _, message1, _, _ = test_conversation
    assert conversation.first_message() == message1


def test_first_message_with_role(test_conversation):
    conversation, _, role_assistant, _, message2, _ = test_conversation
    assert conversation.first_message(role_assistant) == message2


def test_first_message_with_nonexistent_role(test_conversation):
    conversation, _, _, _, _, _ = test_conversation
    role_nonexistent = Role.TOOL
    assert conversation.first_message(role_nonexistent) is None


def test_last_message_no_role(test_conversation):
    conversation, _, _, _, _, message3 = test_conversation
    assert conversation.last_message() == message3


def test_last_message_with_role(test_conversation):
    conversation, role_user, _, _, _, message3 = test_conversation
    assert conversation.last_message(role_user) == message3


def test_last_message_with_nonexistent_role(test_conversation):
    conversation, _, _, _, _, _ = test_conversation
    role_nonexistent = Role.TOOL
    assert conversation.last_message(role_nonexistent) is None


def test_filter_messages_no_role(test_conversation):
    conversation, _, _, message1, message2, message3 = test_conversation
    assert conversation.filter_messages() == [message1, message2, message3]


def test_filter_messages_with_role(test_conversation):
    conversation, role_user, _, message1, _, message3 = test_conversation
    assert conversation.filter_messages(role=role_user) == [message1, message3]
    assert conversation.filter_messages(filter_fn=lambda m: m.role == role_user) == [
        message1,
        message3,
    ]


def test_filter_messages_with_filter_fn(test_conversation):
    conversation, role_user, _, message1, message2, message3 = test_conversation
    assert conversation.filter_messages(filter_fn=lambda m: m.role == role_user) == [
        message1,
        message3,
    ]
    assert conversation.filter_messages(filter_fn=lambda m: m.role != role_user) == [
        message2,
    ]


def test_filter_messages_with_role_and_filter_fn(test_conversation):
    conversation, role_user, _, message1, message2, message3 = test_conversation
    assert conversation.filter_messages(
        role=role_user, filter_fn=lambda m: m.role == role_user
    ) == [
        message1,
        message3,
    ]
    assert (
        conversation.filter_messages(
            role=role_user, filter_fn=lambda m: m.role != role_user
        )
        == []
    )


def test_filter_messages_with_nonexistent_role(test_conversation):
    conversation, _, _, _, _, _ = test_conversation
    role_nonexistent = Role.TOOL
    assert conversation.filter_messages(role=role_nonexistent) == []
    assert (
        conversation.filter_messages(role=role_nonexistent, filter_fn=lambda m: True)
        == []
    )
    assert (
        conversation.filter_messages(filter_fn=lambda m: m.role == role_nonexistent)
        == []
    )


def test_repr(test_conversation):
    conversation, _, _, message1, message2, message3 = test_conversation
    assert repr(message1) == "1 - USER: Hello"
    assert repr(message2) == "ASSISTANT: Hi, how can I help you?"
    assert repr(message3) == "USER: I need assistance with my account. | <IMAGE_BINARY>"
    assert repr(conversation) == (
        "1 - USER: Hello\n"
        "ASSISTANT: Hi, how can I help you?\n"
        "USER: I need assistance with my account. | <IMAGE_BINARY>"
    )


def test_conversation_to_dict_legacy():
    conv = Conversation(
        messages=[
            Message(role=Role.USER, content="Hello"),
            Message(role=Role.ASSISTANT, content="Hi there!"),
        ],
        metadata={"test": "metadata"},
    )
    conv_dict = conv.to_dict()

    assert isinstance(conv_dict, dict)
    assert "messages" in conv_dict
    assert len(conv_dict["messages"]) == 2
    assert conv_dict["metadata"] == {"test": "metadata"}
    assert conv_dict["messages"][0]["role"] == "user"
    assert conv_dict["messages"][0]["content"] == "Hello"
    assert conv_dict["messages"][1]["role"] == "assistant"
    assert conv_dict["messages"][1]["content"] == "Hi there!"


def test_conversation_to_dict_compound_text_content():
    conv = Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[ContentItem(type=Type.TEXT, content="Hello")],
            ),
            Message(
                role=Role.ASSISTANT,
                content=[ContentItem(type=Type.TEXT, content="Hi there!")],
            ),
        ],
        metadata={"test": "metadata"},
    )
    conv_dict = conv.to_dict()

    assert isinstance(conv_dict, dict)
    assert "messages" in conv_dict
    assert len(conv_dict["messages"]) == 2
    assert conv_dict["metadata"] == {"test": "metadata"}
    assert conv_dict["messages"][0]["role"] == "user"
    assert conv_dict["messages"][0]["content"] == [{"content": "Hello", "type": "text"}]
    assert conv_dict["messages"][1]["role"] == "assistant"
    assert conv_dict["messages"][1]["content"] == [
        {"content": "Hi there!", "type": "text"}
    ]


def test_conversation_to_dict_compound_mixed_content():
    png_bytes = _create_test_image_bytes()
    conv = Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(type=Type.IMAGE_BINARY, binary=png_bytes),
                    ContentItem(type=Type.TEXT, content="Hello"),
                ],
            ),
            Message(
                role=Role.ASSISTANT,
                content=[
                    ContentItem(type=Type.TEXT, content="Hi there!"),
                    ContentItem(
                        type=Type.IMAGE_URL,
                        content="/tmp/foo.png",
                        binary=png_bytes,
                    ),
                ],
            ),
        ],
        metadata={"test": "metadata"},
    )
    conv_dict = conv.to_dict()

    assert isinstance(conv_dict, dict)
    assert "messages" in conv_dict
    assert len(conv_dict["messages"]) == 2
    assert conv_dict["metadata"] == {"test": "metadata"}
    assert conv_dict["messages"][0]["role"] == "user"
    assert conv_dict["messages"][0]["content"] == [
        {
            "binary": _SMALL_B64_IMAGE,
            "type": "image_binary",
        },
        {"content": "Hello", "type": "text"},
    ]
    assert conv_dict["messages"][1]["role"] == "assistant"
    assert conv_dict["messages"][1]["content"] == [
        {"content": "Hi there!", "type": "text"},
        {
            "binary": _SMALL_B64_IMAGE,
            "content": "/tmp/foo.png",
            "type": "image_url",
        },
    ]


def test_conversation_from_dict_legacy():
    conv_dict = {
        "messages": [
            {"role": "user", "content": "Hello"},
            {"role": "assistant", "content": "Hi there!"},
        ],
        "metadata": {"test": "metadata"},
    }
    conv = Conversation.from_dict(conv_dict)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert conv.messages[0].content == "Hello"
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_conversation_from_dict_with_unknown_fields():
    conv_dict = {
        "messages": [
            {"role": "user", "content": "Hello", "foo_unknown": "bar"},
            {"role": "assistant", "content": "Hi there!", "type": "text"},
        ],
        "metadata": {"test": "metadata"},
    }
    conv = Conversation.from_dict(conv_dict)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert conv.messages[0].content == "Hello"
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_conversation_from_dict_compound_mixed_content():
    png_bytes = _create_test_image_bytes()
    conv_dict = {
        "messages": [
            {
                "role": "user",
                "content": [
                    {
                        "binary": _SMALL_B64_IMAGE,
                        "type": "image_binary",
                    },
                    {"content": "Hello", "type": "text"},
                ],
            },
            {"role": "assistant", "content": "Hi there!"},
        ],
        "metadata": {"test": "metadata"},
    }
    conv = Conversation.from_dict(conv_dict)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert isinstance(conv.messages[0].content, list)
    assert conv.messages[0].content == [
        ContentItem(type=Type.IMAGE_BINARY, binary=png_bytes),
        ContentItem(type=Type.TEXT, content="Hello"),
    ]
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_conversation_to_json_legacy():
    conv = Conversation(
        messages=[
            Message(role=Role.USER, content="Hello"),
            Message(role=Role.ASSISTANT, content="Hi there!"),
        ],
        metadata={"test": "metadata"},
    )
    json_str = conv.to_json()

    assert isinstance(json_str, str)
    assert '"role":"user"' in json_str
    assert '"content":"Hello"' in json_str
    assert '"role":"assistant"' in json_str
    assert '"content":"Hi there!"' in json_str
    assert '"test":"metadata"' in json_str


def test_conversation_to_json_mixed_content():
    png_bytes = _create_test_image_bytes()
    conv = Conversation(
        messages=[
            Message(
                role=Role.USER,
                content=[
                    ContentItem(type=Type.IMAGE_BINARY, binary=png_bytes),
                    ContentItem(type=Type.TEXT, content="Hello"),
                ],
            ),
            Message(
                role=Role.ASSISTANT,
                content=[
                    ContentItem(type=Type.TEXT, content="Hi there!"),
                    ContentItem(
                        type=Type.IMAGE_URL,
                        content="/tmp/foo.png",
                        binary=png_bytes,
                    ),
                ],
            ),
        ],
        metadata={"test": "_MY_METADATA_"},
    )
    json_str = conv.to_json()

    assert isinstance(json_str, str)
    assert '"role":"user"' in json_str, json_str
    assert '"type":"image_binary"' in json_str
    assert ('"binary":"' + _SMALL_B64_IMAGE + '"') in json_str, json_str
    assert json_str.count('"binary":"' + _SMALL_B64_IMAGE + '"') == 2, json_str
    assert '"content":"Hello"' in json_str, json_str
    assert '"type":"text"' in json_str
    assert json_str.count('"type":"text"') == 2, json_str
    assert '"role":"assistant"' in json_str
    assert '"type":"image_url"' in json_str
    assert '"content":"Hi there!"' in json_str
    assert '"test":"_MY_METADATA_"' in json_str


def test_conversation_from_json_legacy():
    json_str = '{"messages": [{"role": "user", "content": "Hello"}, {"role": "assistant", "content": "Hi there!"}], "metadata": {"test": "metadata"}}'  # noqa: E501
    conv = Conversation.from_json(json_str)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert conv.messages[0].content == "Hello"
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_conversation_from_json_with_unknown_fields():
    json_str = '{"messages": [{"role": "user", "content": "Hello", "foo_unknown": "Z"}, {"role": "assistant", "content": "Hi there!"}], "metadata": {"test": "metadata"}}'  # noqa: E501
    conv = Conversation.from_json(json_str)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert conv.messages[0].content == "Hello"
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_conversation_from_json_compound_simple():
    json_str = '{"messages": [{"role": "user", "content": [{"type": "text", "content": "Hello"}]}, {"role": "assistant", "content": "Hi there!"}], "metadata": {"test": "metadata"}}'  # noqa: E501
    conv = Conversation.from_json(json_str)

    assert isinstance(conv, Conversation)
    assert len(conv.messages) == 2
    assert conv.metadata == {"test": "metadata"}
    assert conv.messages[0].role == Role.USER
    assert isinstance(conv.messages[0].content, list)
    assert conv.messages[0].content == [ContentItem(type=Type.TEXT, content="Hello")]
    assert conv.messages[1].role == Role.ASSISTANT
    assert conv.messages[1].content == "Hi there!"


def test_roundtrip_dict_legacy(root_testdata_dir):
    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(id="", role=Role.ASSISTANT, content="Hi there!"),
            Message(
                role=Role.TOOL,
                content="lalala",
            ),
            Message(
                id="xyz",
                role=Role.USER,
                content="oumi_logo_dark",
            ),
        ],
        metadata={"test": "metadata"},
    )
    conv_dict = original.to_dict()
    reconstructed = Conversation.from_dict(conv_dict)

    assert original == reconstructed


def test_roundtrip_dict_compound_mixed_content(root_testdata_dir):
    png_logo_bytes = load_image_png_bytes_from_path(
        root_testdata_dir / "images" / "oumi_logo_dark.png"
    )
    png_small_image_bytes = _create_test_image_bytes()

    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(id="", role=Role.ASSISTANT, content="Hi there!"),
            Message(
                id="z072",
                role=Role.USER,
                content=[
                    ContentItem(binary=png_logo_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(binary=png_small_image_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(
                        content="https://www.oumi.ai/logo.png",
                        type=Type.IMAGE_URL,
                    ),
                ],
            ),
            Message(
                id="_xyz",
                role=Role.TOOL,
                content=[
                    ContentItem(
                        content=str(
                            root_testdata_dir / "images" / "oumi_logo_dark.png"
                        ),
                        binary=png_logo_bytes,
                        type=Type.IMAGE_PATH,
                    ),
                    ContentItem(
                        content="http://oumi.ai/bzz.png",
                        binary=png_small_image_bytes,
                        type=Type.IMAGE_URL,
                    ),
                    ContentItem(content="<@>", type=Type.TEXT),
                ],
            ),
        ],
        metadata={"a": "b", "b": "c"},
    )
    conv_dict = original.to_dict()
    reconstructed = Conversation.from_dict(conv_dict)

    assert original == reconstructed


def test_roundtrip_json_legacy(root_testdata_dir):
    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(id="", role=Role.ASSISTANT, content="Hi there!"),
            Message(
                role=Role.USER,
                content="",
            ),
            Message(
                id="xyz",
                role=Role.TOOL,
                content="oumi_logo_dark",
            ),
        ],
        metadata={"test": "metadata"},
    )
    json_str = original.to_json()
    reconstructed = Conversation.from_json(json_str)

    assert original == reconstructed


def test_roundtrip_json_compound_mixed_content(root_testdata_dir):
    png_logo_bytes = load_image_png_bytes_from_path(
        root_testdata_dir / "images" / "oumi_logo_light.png"
    )
    png_small_image_bytes = _create_test_image_bytes()

    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(id="", role=Role.ASSISTANT, content="Hi there!"),
            Message(
                id="z072",
                role=Role.USER,
                content=[
                    ContentItem(binary=png_logo_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(binary=png_small_image_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(
                        content="https://www.oumi.ai/logo.png",
                        type=Type.IMAGE_URL,
                    ),
                ],
            ),
            Message(
                id="_xyz",
                role=Role.TOOL,
                content=[
                    ContentItem(
                        content=str(
                            root_testdata_dir / "images" / "oumi_logo_dark.png"
                        ),
                        binary=png_logo_bytes,
                        type=Type.IMAGE_PATH,
                    ),
                    ContentItem(
                        content="http://oumi.ai/bzz.png",
                        binary=png_small_image_bytes,
                        type=Type.IMAGE_URL,
                    ),
                    ContentItem(content="<@>", type=Type.TEXT),
                ],
            ),
        ],
        metadata={"a": "b", "b": "c"},
    )
    json_str = original.to_json()
    reconstructed = Conversation.from_json(json_str)

    assert original == reconstructed


def test_roundtrip_proto_legacy(root_testdata_dir):
    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(role=Role.ASSISTANT, content="Hi there!"),
            Message(
                role=Role.USER,
                content="",
            ),
            Message(
                id="xyz",
                role=Role.TOOL,
                content="oumi_logo_dark",
            ),
        ],
        metadata={"test": "metadata"},
    )
    convo_proto = original.to_proto()
    reconstructed = Conversation.from_proto(convo_proto)

    assert original == reconstructed, (
        f"\n\noriginal: {original.to_dict()}\n"
        f"\n\nreconstructed: {reconstructed.to_dict()}\n"
        f"\n\nproto: {convo_proto.SerializeToString()}"
    )


def test_roundtrip_proto_compound_mixed_content(root_testdata_dir):
    png_logo_bytes = load_image_png_bytes_from_path(
        root_testdata_dir / "images" / "oumi_logo_light.png"
    )
    png_small_image_bytes = _create_test_image_bytes()

    original = Conversation(
        messages=[
            Message(id="001", role=Role.SYSTEM, content="Behave!"),
            Message(role=Role.ASSISTANT, content="Hi there!"),
            Message(
                id="z072",
                role=Role.USER,
                content=[
                    ContentItem(binary=png_logo_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(binary=png_small_image_bytes, type=Type.IMAGE_BINARY),
                    ContentItem(
                        content="https://www.oumi.ai/logo.png",
                        type=Type.IMAGE_URL,
                    ),
                ],
            ),
            Message(
                id="_xyz",
                role=Role.TOOL,
                content=[
                    ContentItem(
                        content=str(
                            root_testdata_dir / "images" / "oumi_logo_dark.png"
                        ),
                        binary=png_logo_bytes,
                        type=Type.IMAGE_PATH,
                    ),
                    ContentItem(
                        content="http://oumi.ai/bzz.png",
                        binary=png_small_image_bytes,
                        type=Type.IMAGE_URL,
                    ),
                    ContentItem(content="<@>", type=Type.TEXT),
                ],
            ),
        ],
        metadata={"a": "b", "b": "c"},
    )
    convo_proto = original.to_proto()
    reconstructed = Conversation.from_proto(convo_proto)

    assert original == reconstructed, (
        f"\n\noriginal: {original.to_dict()}\n"
        f"\n\nreconstructed: {reconstructed.to_dict()}\n"
        f"\n\nproto: {convo_proto.SerializeToString()}"
    )


def test_from_dict_with_invalid_field():
    with pytest.raises(ValueError, match="Field required"):
        Conversation.from_dict({"invalid": "data"})


def test_from_json_with_invalid_field():
    with pytest.raises(ValueError, match="Invalid JSON"):
        Conversation.from_json('{"invalid": json')


def test_from_dict_with_invalid_base64():
    with pytest.raises(ValueError, match="Invalid base64-encoded string"):
        Conversation.from_dict(
            {
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            {
                                "binary": "INVALID_BASE64!",
                                "type": "image_binary",
                            }
                        ],
                    },
                ],
                "metadata": {"test": "metadatazzz"},
            }
        )


def test_empty_content_list():
    message = Message(role=Role.ASSISTANT, content=[])
    assert isinstance(message.content, list)
    assert len(message.content) == 0


def test_empty_content_string():
    message = Message(role=Role.USER, content="")
    assert isinstance(message.content, str)
    assert len(message.content) == 0


def test_incorrect_message_content_item_type():
    with pytest.raises(ValueError, match="Input should be a valid string"):
        ContentItem(
            type=Type.TEXT,
            content=cast(str, 12345.7),  # Hacky way to pass a number as content.
        )
    with pytest.raises(ValueError, match="Either content or binary must be provided"):
        ContentItem(
            type=Type.TEXT,
        )
    with pytest.raises(ValueError, match="No image bytes in message content item"):
        ContentItem(type=Type.IMAGE_BINARY, binary=b"")
    with pytest.raises(ValueError, match="Content not provided"):
        ContentItem(type=Type.IMAGE_URL, binary=b"")
    with pytest.raises(ValueError, match="Content not provided"):
        ContentItem(type=Type.IMAGE_PATH, binary=b"")
    with pytest.raises(ValueError, match="Binary can only be provided for images"):
        ContentItem(type=Type.TEXT, binary=b"")


@pytest.mark.parametrize(
    "role",
    [Role.USER, Role.ASSISTANT, Role.TOOL, Role.SYSTEM],
)
def test_content_item_methods_mixed_items(role: Role):
    text_item1 = ContentItem(type=Type.TEXT, content="aaa")
    image_item1 = ContentItem(type=Type.IMAGE_BINARY, binary=_create_test_image_bytes())
    text_item2 = ContentItem(type=Type.TEXT, content=" B B ")
    image_item2 = ContentItem(
        type=Type.IMAGE_PATH,
        content="/tmp/test/dummy.jpeg",
        binary=_create_test_image_bytes(),
    )
    text_item3 = ContentItem(type=Type.TEXT, content="CC")

    message = Message(
        role=role,
        content=[
            text_item1,
            image_item1,
            text_item2,
            image_item2,
            text_item3,
        ],
    )

    assert message.contains_text()
    assert not message.contains_single_text_content_item_only()
    assert not message.contains_text_content_items_only()

    assert message.contains_images()
    assert not message.contains_single_image_content_item_only()
    assert not message.contains_image_content_items_only()

    assert message.compute_flattened_text_content() == "aaa  B B  CC"
    assert message.compute_flattened_text_content("||") == "aaa|| B B ||CC"

    assert message.content_items == [
        text_item1,
        image_item1,
        text_item2,
        image_item2,
        text_item3,
    ]
    assert message.image_content_items == [image_item1, image_item2]
    assert message.text_content_items == [text_item1, text_item2, text_item3]

    assert message.count_content_items() == ContentItemCounts(
        total_items=5, image_items=2, text_items=3
    )


@pytest.mark.parametrize(
    "image_type",
    [Type.IMAGE_BINARY, Type.IMAGE_PATH, Type.IMAGE_URL],
)
def test_content_item_methods_single_image(image_type):
    test_image_item = ContentItem(
        type=image_type,
        content=(None if image_type == Type.IMAGE_BINARY else "foo"),
        binary=(
            _create_test_image_bytes() if image_type == Type.IMAGE_BINARY else None
        ),
    )
    message = Message(
        role=Role.ASSISTANT,
        content=[test_image_item],
    )

    assert not message.contains_text()
    assert not message.contains_single_text_content_item_only()
    assert not message.contains_text_content_items_only()

    assert message.contains_images()
    assert message.contains_single_image_content_item_only()
    assert message.contains_image_content_items_only()

    assert message.compute_flattened_text_content() == ""
    assert message.compute_flattened_text_content("Z") == ""

    assert message.content_items == [
        test_image_item,
    ]
    assert message.image_content_items == [test_image_item]
    assert message.text_content_items == []

    assert message.count_content_items() == ContentItemCounts(
        total_items=1, image_items=1, text_items=0
    )


def test_content_item_methods_triple_image():
    test_image_item1 = ContentItem(
        type=Type.IMAGE_BINARY,
        binary=(_create_test_image_bytes()),
    )
    test_image_item2 = ContentItem(
        type=Type.IMAGE_URL,
        content="http://oumi.ai/a.png",
    )
    test_image_item3 = ContentItem(
        type=Type.IMAGE_PATH,
        content="/tmp/oumi.ai/b.gif",
    )
    message = Message(
        role=Role.ASSISTANT,
        content=[test_image_item1, test_image_item2, test_image_item3],
    )

    assert not message.contains_text()
    assert not message.contains_single_text_content_item_only()
    assert not message.contains_text_content_items_only()

    assert message.contains_images()
    assert not message.contains_single_image_content_item_only()
    assert message.contains_image_content_items_only()

    assert message.compute_flattened_text_content() == ""
    assert message.compute_flattened_text_content("Z") == ""

    assert message.content_items == [
        test_image_item1,
        test_image_item2,
        test_image_item3,
    ]
    assert message.image_content_items == [
        test_image_item1,
        test_image_item2,
        test_image_item3,
    ]
    assert message.text_content_items == []

    assert message.count_content_items() == ContentItemCounts(
        total_items=3, image_items=3, text_items=0
    )


def test_content_item_methods_legacy_text():
    test_text_item = ContentItem(type=Type.TEXT, content="bzzz")
    message = Message(
        role=Role.USER,
        content=(test_text_item.content or ""),
    )

    assert message.contains_text()
    assert message.contains_single_text_content_item_only()
    assert message.contains_text_content_items_only()

    assert not message.contains_images()
    assert not message.contains_single_image_content_item_only()
    assert not message.contains_image_content_items_only()

    assert message.compute_flattened_text_content() == "bzzz"
    assert message.compute_flattened_text_content("X") == "bzzz"

    assert message.content_items == [
        test_text_item,
    ]
    assert message.image_content_items == []
    assert message.text_content_items == [test_text_item]

    assert message.count_content_items() == ContentItemCounts(
        total_items=1, image_items=0, text_items=1
    )


def test_content_item_methods_double_text():
    test_text_item = ContentItem(type=Type.TEXT, content="bzzz")
    message = Message(
        role=Role.USER,
        content=[test_text_item, test_text_item],
    )

    assert message.contains_text()
    assert not message.contains_single_text_content_item_only()
    assert message.contains_text_content_items_only()

    assert not message.contains_images()
    assert not message.contains_single_image_content_item_only()
    assert not message.contains_image_content_items_only()

    assert message.compute_flattened_text_content() == "bzzz bzzz"
    assert message.compute_flattened_text_content("^") == "bzzz^bzzz"

    assert message.content_items == [
        test_text_item,
        test_text_item,
    ]
    assert message.image_content_items == []
    assert message.text_content_items == [test_text_item, test_text_item]

    assert message.count_content_items() == ContentItemCounts(
        total_items=2, image_items=0, text_items=2
    )


def test_role_str_repr():
    assert str(Role.ASSISTANT) == "assistant"
    assert "assistant" in repr(Role.ASSISTANT)
    assert str(Role.USER) == "user"
    assert "user" in repr(Role.USER)
    assert str(Role.TOOL) == "tool"
    assert "tool" in repr(Role.TOOL)


def test_type_str_repr():
    assert str(Type.TEXT) == "text"
    assert "text" in repr(Type.TEXT)
    assert str(Type.IMAGE_BINARY) == "image_binary"
    assert "image_binary" in repr(Type.IMAGE_BINARY)
    assert str(Type.IMAGE_URL) == "image_url"
    assert "image_url" in repr(Type.IMAGE_URL)
    assert str(Type.IMAGE_PATH) == "image_path"
    assert "image_path" in repr(Type.IMAGE_PATH)


def test_frozen_message_content_item():
    test_item = ContentItem(type=Type.TEXT, content="init")
    with pytest.raises(pydantic.ValidationError, match="Instance is frozen"):
        test_item.content = "foo"
    assert test_item.content == "init"

    with pytest.raises(pydantic.ValidationError, match="Instance is frozen"):
        test_item.type = Type.IMAGE_BINARY
    assert test_item.type == Type.TEXT


def test_frozen_message():
    test_item = ContentItem(type=Type.TEXT, content="bzzz")
    message = Message(
        id="007",
        role=Role.ASSISTANT,
        content=[test_item, test_item],
    )

    with pytest.raises(pydantic.ValidationError, match="Instance is frozen"):
        message.id = "001"
    assert message.id == "007"

    with pytest.raises(pydantic.ValidationError, match="Instance is frozen"):
        message.role = Role.TOOL
    assert message.role == Role.ASSISTANT

    with pytest.raises(pydantic.ValidationError, match="Instance is frozen"):
        message.content = "Hey"
    assert isinstance(message.content, list)
    assert len(message.content) == 2

    # Pydantic "frozen" only ensures that `message.content` can't be re-assigned
    # but it doesn't enforce the field object itself is immutable.
    message.content.append(test_item)
    assert isinstance(message.content, list)
    assert len(message.content) == 3



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import Any, Optional

from oumi.core.configs.params.base_params import BaseParams
from oumi.core.configs.params.guided_decoding_params import GuidedDecodingParams


@dataclass
class GenerationParams(BaseParams):
    max_new_tokens: int = 1024
    """The maximum number of new tokens to generate.

    This limits the length of the generated text to prevent excessively long outputs.
    Default is 1024 tokens.
    """

    batch_size: Optional[int] = 1
    """The number of sequences to generate in parallel.

    Larger batch sizes can improve throughput but require more memory. Default is 1.

    The value must either be positive or None, in which case the behavior is dependent
    on the downstream application. For example, LM Harness will automatically determine
    the largest batch size that will fit in memory.

    For inference, this parameter is only used in `NativeTextInferenceEngine`.
    """

    exclude_prompt_from_response: bool = True
    """Whether to trim the model's response and remove the prepended prompt."""

    seed: Optional[int] = None
    """Seed to use for random number determinism.
    If specified, APIs may use this parameter to make a best-effort at determinism.
    """

    temperature: float = 0.0
    """Controls randomness in the output.

    Higher values (e.g., 1.0) make output more random, while lower values (e.g., 0.2)
    make it more focused and deterministic.
    """

    top_p: float = 1.0
    """An alternative to temperature, called nucleus sampling.

    It sets the cumulative probability threshold for token selection.
    For example, 0.9 means only considering the tokens comprising
    the top 90% probability mass.
    """

    frequency_penalty: float = 0.0
    """Positive values penalize new tokens based on their existing frequency in the text
    so far, decreasing the model's likelihood to repeat the same line verbatim.
    """

    presence_penalty: float = 0.0
    """Positive values penalize new tokens based on whether they appear in the text
    so far, increasing the model's likelihood to talk about new topics.
    """

    stop_strings: Optional[list[str]] = None
    """List of sequences where the API will stop generating further tokens."""

    stop_token_ids: Optional[list[int]] = None
    """List of token ids for which the API will stop generating further tokens. This
    is only supported in `VLLMInferenceEngine` and `NativeTextInferenceEngine`."""

    logit_bias: dict[Any, float] = field(default_factory=dict)
    """Modify the likelihood of specified tokens appearing in the completion.

    Keys are tokens (specified by their token ID in the tokenizer),
    and values are the bias (-100 to 100). Mathematically, the bias is added
    to the logits generated by the model prior to sampling. The exact effect will
    vary per model, but values between -1 and 1 should decrease or increase
    likelihood of selection; values like -100 or 100 should result in a ban or
    exclusive selection of the relevant token.
    """

    min_p: float = 0.0
    """Sets a minimum probability threshold for token selection.

    Tokens with probabilities below this threshold are filtered out before top-p
    or top-k sampling. This can help prevent the selection of highly improbable tokens.
    Default is 0.0 (no minimum threshold).
    """

    use_cache: bool = False
    """Whether to use the model's internal cache (key/value attentions) to speed up
    generation.
    Default is False.
    """

    num_beams: int = 1
    """Number of beams for beam search. 1 means no beam search. Larger number of beams
    will make for a more thorough search for probable output token sequences, at
    the cost of increased computation time.
    Default is 1.
    """

    use_sampling: bool = False
    """Whether to use sampling for next-token generation. If False, uses greedy
    decoding.
    Default is False."""

    guided_decoding: Optional[GuidedDecodingParams] = None
    """Parameters for guided decoding."""

    skip_special_tokens: bool = True
    """Whether to skip special tokens when decoding the generated text.

    When True (default), special tokens like <eos>, <pad>, etc. are removed from
    the output text. When False, these tokens are included in the decoded text.
    This can be useful for models that use special tokens as part of their output
    format (e.g., reasoning tokens, tool call markers).
    """

    def __post_init__(self):
        """Validates generation-specific parameters."""
        if self.batch_size is not None and self.batch_size < 1:
            raise ValueError("Batch size must be positive.")

        if self.num_beams < 1:
            raise ValueError("num_beams must be strictly larger than 0.")

        if self.temperature < 0:
            raise ValueError("Temperature must be non-negative.")

        if not 0 <= self.top_p <= 1:
            raise ValueError("top_p must be between 0 and 1.")

        for token_id, bias in self.logit_bias.items():
            if not isinstance(token_id, (str, int)):
                raise ValueError(
                    f"Logit bias token ID {token_id} must be an integer or a string."
                )

            if not -100 <= bias <= 100:
                raise ValueError(
                    f"Logit bias for token {token_id} must be between -100 and 100."
                )

        if not 0 <= self.min_p <= 1:
            raise ValueError("min_p must be between 0 and 1.")



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Core trainers module for the Oumi (Open Universal Machine Intelligence) library.

This module provides various trainer implementations for use in the Oumi framework.
These trainers are designed to facilitate the training process for different
types of models and tasks.

Example:
    >>> from oumi.core.trainers import Trainer
    >>> trainer = Trainer(model=my_model, dataset=my_dataset) # doctest: +SKIP
    >>> trainer.train() # doctest: +SKIP

Note:
    For detailed information on each trainer, please refer to their respective
        class documentation.
"""

from oumi.core.trainers.base_trainer import BaseTrainer
from oumi.core.trainers.hf_trainer import HuggingFaceTrainer
from oumi.core.trainers.oumi_trainer import Trainer
from oumi.core.trainers.trl_dpo_trainer import TrlDpoTrainer
from oumi.core.trainers.verl_grpo_trainer import VerlGrpoTrainer

__all__ = [
    "BaseTrainer",
    "HuggingFaceTrainer",
    "Trainer",
    "TrlDpoTrainer",
    "VerlGrpoTrainer",
]



# Copyright 2025 - Oumi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import transformers

from oumi.core.constants import LABEL_IGNORE_INDEX
from oumi.core.tokenizers.base_tokenizer import BaseTokenizer
from oumi.core.types import Conversation
from oumi.utils.logging import logger


#
# Base class functions
#
def tokenize_for_completions_only_training_with_template(
    tokenizer: BaseTokenizer, conversation: Conversation
) -> dict:
    """Tokenize a conversation for completions-only training with a template."""
    batch: transformers.BatchEncoding = tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=True,
        return_dict=True,
        return_assistant_tokens_mask=True,
    )

    data = batch.data

    assistant_tokens_mask = data.pop("assistant_masks")

    data["labels"] = [
        token_id if mask else LABEL_IGNORE_INDEX
        for mask, token_id in zip(assistant_tokens_mask, data["input_ids"])
    ]

    return data


def tokenize_for_completions_only_training_with_prefix(
    tokenizer: BaseTokenizer,
    conversation: Conversation,
    response_template: str,
    instruction_template: str,
    response_token_ids: list[int],
    instruction_token_ids: list[int],
) -> dict:
    """Tokenize a conversation for completions-only training with a prefix."""
    prompt: str = tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=False,
        return_dict=False,
        return_assistant_tokens_mask=False,
    )
    tokenizer_batch: transformers.BatchEncoding = tokenizer(
        prompt, truncation=True, padding=False, return_tensors="pt"
    )

    batch = {k: v[0] for k, v in tokenizer_batch.data.items()}
    batch["labels"] = batch["input_ids"].clone()

    response_token_ids_idxs = []
    human_token_ids_idxs = []

    cond = np.atleast_1d(batch["labels"] == response_token_ids[0])

    for assistant_idx in np.where(cond)[0]:
        # find the indexes of the start of a response.
        if (
            response_token_ids
            == batch["labels"][
                assistant_idx : assistant_idx + len(response_token_ids)
            ].tolist()
        ):
            response_token_ids_idxs.append(assistant_idx + len(response_token_ids))

    if len(response_token_ids_idxs) == 0:
        logger.warning(
            f"Could not find response key `{response_template}` in the "
            f"following instance: {tokenizer.decode(batch['input_ids'])} "
            f"This instance will be ignored in loss calculation. "
            f"Note, if this happens often, consider increasing the `max_seq_length`."
        )
        batch["labels"][:] = LABEL_IGNORE_INDEX

    human_token_ids = instruction_token_ids
    for human_idx in np.where(batch["labels"] == human_token_ids[0])[0]:
        # find the indexes of the start of a human answer.
        if (
            human_token_ids
            == batch["labels"][human_idx : human_idx + len(human_token_ids)].tolist()
        ):
            human_token_ids_idxs.append(human_idx)

    if len(human_token_ids_idxs) == 0:
        logger.warning(
            f"Could not find instruction key `{instruction_template}` in the "
            f"following instance: {tokenizer.decode(batch['input_ids'])} "
            f"This instance will be ignored in loss calculation. "
            f"Note, if this happens often, consider increasing the `max_seq_length`."
        )
        batch["labels"][:] = LABEL_IGNORE_INDEX

    if (
        len(human_token_ids_idxs) > 0
        and len(response_token_ids_idxs) > 0
        and human_token_ids_idxs[0] > response_token_ids_idxs[0]
    ):
        human_token_ids_idxs = [0] + human_token_ids_idxs

    for idx, (start, end) in enumerate(
        zip(human_token_ids_idxs, response_token_ids_idxs)
    ):
        # Make pytorch loss function ignore all non response tokens
        if idx != 0:
            batch["labels"][start:end] = LABEL_IGNORE_INDEX
        else:
            batch["labels"][:end] = LABEL_IGNORE_INDEX

    if len(response_token_ids_idxs) < len(human_token_ids_idxs):
        batch["labels"][human_token_ids_idxs[-1] :] = LABEL_IGNORE_INDEX

    return batch


#
# Multi-turn collator functions
#
def mask_labels_without_user_template(
    labels: np.ndarray,
    response_token_ids: list[int],
    ignore_index: int = LABEL_IGNORE_INDEX,
) -> None:
    """Apply completion-only masking when no user template is provided.

    This strategy masks everything except the last assistant response, allowing
    the model to learn only from the final assistant turn in the conversation.

    Args:
        labels: Label array to mask
        response_token_ids: Token IDs of the response template.
        ignore_index: Value to use for masked positions
    """
    # Find all response positions
    response_starts = find_all_sequences(labels, response_token_ids)

    if not response_starts:
        # No assistant responses found, mask everything
        labels[:] = ignore_index
        return

    # Save original labels before masking
    original_labels = labels.copy()

    # Mask everything initially
    labels[:] = ignore_index

    # Only unmask the last assistant response
    last_response_start = response_starts[-1]

    # Unmask from the last response start to the end of the sequence
    labels[last_response_start:] = original_labels[last_response_start:]


def mask_labels_for_completions_only(
    labels: np.ndarray,
    response_token_ids: list[int],
    instruction_token_ids: list[int],
    ignore_index: int = LABEL_IGNORE_INDEX,
) -> None:
    """Apply completion-only masking to labels with user and assistant templates.

    This strategy masks everything except assistant response content, using user
    templates to determine the boundaries of each assistant response.

    Args:
        labels: Label array to mask
        response_token_ids: Token IDs of the response template.
        instruction_token_ids: Token IDs of the instruction template.
        ignore_index: Value to use for masked positions
    """
    # Find all response and user positions
    response_starts = find_all_sequences(labels, response_token_ids)
    user_starts = find_all_sequences(labels, instruction_token_ids)

    # If no response templates found, mask everything
    if not response_starts:
        labels[:] = ignore_index
        return

    # Save original labels before masking
    original_labels = labels.copy()

    # Mask everything except assistant responses
    labels[:] = ignore_index  # Start by masking everything

    # Unmask each assistant response (content after the template)
    for resp_start in response_starts:
        # Find the next user template start after this response
        resp_end = len(labels)  # Default to end of sequence
        for user_start in user_starts:
            # user_start is position after user template, so we need to go back
            user_template_start = user_start - len(instruction_token_ids)
            if user_template_start > resp_start:
                resp_end = user_template_start
                break

        # Restore the original labels for the response content only
        # (starting after the response template)
        labels[resp_start:resp_end] = original_labels[resp_start:resp_end]


def find_all_sequences(arr: np.ndarray, target: list[int]) -> list[int]:
    """Find all occurrences of target sequence in array.

    Returns the positions of the target sequence AFTER the found sequence.
    """
    arr_list = arr.tolist()
    positions = []
    for i in range(len(arr_list) - len(target) + 1):
        if arr_list[i : i + len(target)] == target:
            positions.append(i + len(target))  # Return position after the sequence
    return positions


#
# Utils
#
def tokenizer_for_inference(
    tokenizer: BaseTokenizer, conversation: Conversation
) -> dict:
    """Tokenize a conversation for inference."""
    return tokenizer.apply_chat_template(
        conversation=conversation,  # type: ignore
        tokenize=True,
        return_dict=True,
    )



#    Copyright 2023 Haotian Liu
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from abc import ABC, abstractmethod

import torch
import torch.nn as nn
import torch.nn.functional as F

from oumi.models.experimental.cambrian.utils import IS_XLA_AVAILABLE
from oumi.utils.logging import logger

from ..constants import (
    DEFAULT_IM_END_TOKEN,
    DEFAULT_IM_START_TOKEN,
    DEFAULT_IMAGE_PATCH_TOKEN,
    IGNORE_INDEX,
    IMAGE_TOKEN_INDEX,
)
from .multimodal_encoder.builder import build_vision_tower_aux_list
from .multimodal_projector.builder import build_vision_projector
from .vision_sampler import VisionTokenSampler


class CambrianMetaModel:
    def __init__(self, config):
        super().__init__(config)

        logger.info(f"Initializing Cambrian Meta model: {type(self)}...")

        if hasattr(config, "mm_vision_tower_aux_list"):
            projector_type = getattr(config, "mm_projector_type", "linear")
            if projector_type == "sva":
                vision_hidden_size = config.vision_hidden_size
                num_query_group = config.num_query_group
                query_num_list = config.query_num_list
                connector_only = config.connector_only
                connector_depth = config.connector_depth
                self.vision_tower_aux_list = build_vision_tower_aux_list(
                    config, delay_load=True
                )
                self.mm_projector = nn.Sequential(
                    nn.Linear(vision_hidden_size * num_query_group, config.hidden_size),
                    nn.GELU(),
                    nn.Linear(config.hidden_size, config.hidden_size),
                )

                image_token_len = config.image_token_len
                vision_tower_aux_token_len_list = (
                    self.config.mm_vision_tower_aux_token_len_list
                )
                cross_att_token_len_list = [
                    int(vision_tower_aux_token_len**0.5) // int(image_token_len**0.5)
                    for vision_tower_aux_token_len in vision_tower_aux_token_len_list
                ]

                for aux_i, vision_tower_aux in enumerate(self.vision_tower_aux_list):
                    setattr(
                        self,
                        f"mm_projector_aux_{aux_i}",
                        nn.Sequential(
                            nn.Linear(vision_tower_aux.hidden_size, vision_hidden_size),
                            nn.GELU(),
                            nn.Linear(vision_hidden_size, vision_hidden_size),
                            nn.LayerNorm(vision_hidden_size),
                        ),
                    )

                for query_group_i in range(num_query_group):
                    cross_att_token_len_list = [
                        int(vision_tower_aux_token_len**0.5)
                        // int(query_num_list[query_group_i] ** 0.5)
                        for vision_tower_aux_token_len in vision_tower_aux_token_len_list
                    ]
                    setattr(
                        self,
                        f"vision_sampler_{query_group_i}",
                        VisionTokenSampler(
                            vision_hidden_size,
                            vision_hidden_size,
                            [vision_hidden_size] * len(self.vision_tower_aux_list),
                            cross_att_token_len_list,
                            vision_hidden_size,
                            connector_depth,
                        ),
                    )

                if not connector_only:
                    num_of_vision_sampler_layers = (
                        config.num_of_vision_sampler_layers
                    ) = config.num_of_vision_sampler_layers
                    config.start_of_vision_sampler_layers = (
                        config.start_of_vision_sampler_layers
                    )
                    config.stride_of_vision_sampler_layers = (
                        config.stride_of_vision_sampler_layers
                    )
                    cross_att_token_len_list = [
                        int(vision_tower_aux_token_len**0.5)
                        // int(image_token_len**0.5)
                        for vision_tower_aux_token_len in vision_tower_aux_token_len_list
                    ]
                    self.vision_sampler_layers = nn.ModuleList(
                        [
                            VisionTokenSampler(
                                config.hidden_size,
                                vision_hidden_size,
                                [vision_hidden_size] * len(self.vision_tower_aux_list),
                                cross_att_token_len_list,
                                vision_hidden_size,
                                1,
                            )
                            for layer_idx in range(0, num_of_vision_sampler_layers)
                        ]
                    )

                self.vision_query = nn.Parameter(
                    torch.randn((num_query_group, vision_hidden_size), dtype=self.dtype)
                )

                self.image_newline = nn.Parameter(
                    torch.empty(config.hidden_size, dtype=self.dtype)
                )

            else:
                self.vision_tower_aux_list = build_vision_tower_aux_list(
                    config, delay_load=True
                )
                config.mm_hidden_size = sum(
                    [
                        vision_tower_aux.hidden_size
                        for vision_tower_aux in self.vision_tower_aux_list
                    ]
                )
                self.mm_projector = build_vision_projector(config)
                self.image_newline = nn.Parameter(
                    torch.empty(config.hidden_size, dtype=self.dtype)
                )

    # def get_vision_tower(self):
    #     vision_tower = getattr(self, 'vision_tower', None)
    #     if type(vision_tower) is list:
    #         vision_tower = vision_tower[0]
    #     return vision_tower

    def get_vision_tower_aux_list(self):
        vision_tower_aux_list = getattr(self, "vision_tower_aux_list", None)
        return vision_tower_aux_list

    def initialize_vision_modules(self, model_args, fsdp=None):
        # vision_tower = model_args.vision_tower
        num_query_group = model_args.num_query_group
        query_num_list = model_args.query_num_list
        vision_hidden_size = model_args.vision_hidden_size
        vision_tower_aux_list = model_args.vision_tower_aux_list
        vision_tower_aux_token_len_list = model_args.vision_tower_aux_token_len_list
        image_token_len = model_args.image_token_len
        mm_vision_select_layer = model_args.mm_vision_select_layer
        mm_vision_select_feature = model_args.mm_vision_select_feature
        pretrain_mm_mlp_adapter = model_args.pretrain_mm_mlp_adapter
        connector_only = model_args.connector_only
        connector_depth = model_args.connector_depth

        # self.config.mm_vision_tower = vision_tower
        self.config.image_token_len = image_token_len
        self.config.num_query_group = num_query_group
        self.config.query_num_list = query_num_list
        assert num_query_group == len(query_num_list)
        self.config.connector_depth = connector_depth
        self.config.mm_vision_tower_aux_list = vision_tower_aux_list
        self.config.mm_vision_tower_aux_token_len_list = vision_tower_aux_token_len_list
        self.config.connector_only = connector_only

        if self.get_vision_tower_aux_list() is None:
            vision_tower_aux_list = build_vision_tower_aux_list(model_args)
            if model_args.unfreeze_mm_vision_tower:
                self.vision_tower_aux_list = nn.ModuleList(vision_tower_aux_list)
            else:
                self.vision_tower_aux_list = vision_tower_aux_list
        else:
            vision_tower_aux_list = self.vision_tower_aux_list
            for vision_tower_aux in vision_tower_aux_list:
                vision_tower_aux.load_model()

        self.config.use_mm_proj = True
        self.config.mm_projector_type = getattr(
            model_args, "mm_projector_type", "linear"
        )
        self.config.vision_hidden_size = vision_hidden_size
        self.config.mm_vision_select_layer = mm_vision_select_layer
        self.config.mm_vision_select_feature = mm_vision_select_feature

        if getattr(self, "mm_projector", None) is None:
            if self.config.mm_projector_type == "sva":
                self.mm_projector = nn.Sequential(
                    nn.Linear(
                        vision_hidden_size * num_query_group, self.config.hidden_size
                    ),
                    nn.GELU(),
                    nn.Linear(self.config.hidden_size, self.config.hidden_size),
                )
                for aux_i, vision_tower_aux in enumerate(vision_tower_aux_list):
                    setattr(
                        self,
                        f"mm_projector_aux_{aux_i}",
                        nn.Sequential(
                            nn.Linear(vision_tower_aux.hidden_size, vision_hidden_size),
                            nn.GELU(),
                            nn.Linear(vision_hidden_size, vision_hidden_size),
                            nn.LayerNorm(vision_hidden_size),
                        ),
                    )

                # vision sampler for each group of query as the connector before the LLM
                for query_group_i in range(num_query_group):
                    cross_att_token_len_list = [
                        int(vision_tower_aux_token_len**0.5)
                        // int(query_num_list[query_group_i] ** 0.5)
                        for vision_tower_aux_token_len in vision_tower_aux_token_len_list
                    ]
                    setattr(
                        self,
                        f"vision_sampler_{query_group_i}",
                        VisionTokenSampler(
                            vision_hidden_size,
                            vision_hidden_size,
                            [vision_hidden_size] * len(vision_tower_aux_list),
                            cross_att_token_len_list,
                            vision_hidden_size,
                            connector_depth,
                        ),
                    )

                # sampler layers within LLM
                if not connector_only:
                    num_of_vision_sampler_layers = (
                        self.config.num_of_vision_sampler_layers
                    ) = model_args.num_of_vision_sampler_layers
                    self.config.start_of_vision_sampler_layers = (
                        model_args.start_of_vision_sampler_layers
                    )
                    self.config.stride_of_vision_sampler_layers = (
                        model_args.stride_of_vision_sampler_layers
                    )
                    cross_att_token_len_list = [
                        int(vision_tower_aux_token_len**0.5)
                        // int(image_token_len**0.5)
                        for vision_tower_aux_token_len in vision_tower_aux_token_len_list
                    ]
                    self.vision_sampler_layers = nn.ModuleList(
                        [
                            VisionTokenSampler(
                                self.config.hidden_size,
                                vision_hidden_size,
                                [vision_hidden_size] * len(vision_tower_aux_list),
                                cross_att_token_len_list,
                                vision_hidden_size,
                                1,
                            )
                            for layer_idx in range(0, num_of_vision_sampler_layers)
                        ]
                    )
                vision_embed_std = 1 / torch.sqrt(
                    torch.tensor(vision_hidden_size, dtype=self.dtype)
                )
                self.vision_query = nn.Parameter(
                    torch.randn((num_query_group, vision_hidden_size), dtype=self.dtype)
                    * vision_embed_std
                )

                embed_std = 1 / torch.sqrt(
                    torch.tensor(self.config.hidden_size, dtype=self.dtype)
                )
                self.image_newline = nn.Parameter(
                    torch.randn(self.config.hidden_size, dtype=self.dtype) * embed_std
                )

            else:
                self.config.mm_hidden_size = sum(
                    [
                        vision_tower_aux.hidden_size
                        for vision_tower_aux in vision_tower_aux_list
                    ]
                )
                self.mm_projector = build_vision_projector(self.config)
                embed_std = 1 / torch.sqrt(
                    torch.tensor(self.config.hidden_size, dtype=self.dtype)
                )
                self.image_newline = nn.Parameter(
                    torch.randn(self.config.hidden_size, dtype=self.dtype) * embed_std
                )
        else:
            # In case it is frozen by LoRA
            for p in self.mm_projector.parameters():
                p.requires_grad = True

        if pretrain_mm_mlp_adapter is not None:
            mm_projector_weights = torch.load(
                pretrain_mm_mlp_adapter, map_location="cpu"
            )

            def get_w(weights, keyword):
                return {
                    k.split(keyword + ".")[1]: v
                    for k, v in weights.items()
                    if keyword + "." in k
                }

            self.mm_projector.load_state_dict(
                get_w(mm_projector_weights, "mm_projector"), strict=True
            )

            if self.config.mm_projector_type == "sva":
                for aux_i in range(len(vision_tower_aux_list)):
                    getattr(self, f"mm_projector_aux_{aux_i}").load_state_dict(
                        get_w(mm_projector_weights, f"mm_projector_aux_{aux_i}"),
                        strict=True,
                    )

                for query_group_i in range(num_query_group):
                    getattr(self, f"vision_sampler_{query_group_i}").load_state_dict(
                        get_w(
                            mm_projector_weights,
                            f"vision_sampler_{query_group_i}",
                        ),
                        strict=True,
                    )

                if not connector_only:
                    self.vision_sampler_layers.load_state_dict(
                        get_w(mm_projector_weights, "vision_sampler_layers"),
                        strict=True,
                    )
                self.vision_query.data = mm_projector_weights["model.vision_query"]
            self.image_newline.data = mm_projector_weights["model.image_newline"]


def unmask_attention_mask(mask, original_size):
    original_w, original_h = original_size
    cur_h, cur_w = mask.shape[1:3]

    original_aspect_ratio = original_w / original_h
    current_aspect_ratio = cur_w / cur_h

    if original_aspect_ratio > current_aspect_ratio:
        scale_factor = cur_w / original_w
        new_height = int(original_h * scale_factor)
        padding = (cur_h - new_height) // 2
        if padding > 0:
            mask[:, :padding, :] = 0
            mask[:, -padding:, :] = 0
        return mask
    else:
        scale_factor = cur_h / original_h
        new_width = int(original_w * scale_factor)
        padding = (cur_w - new_width) // 2
        if padding > 0:
            mask[:, :, :padding] = 0
            mask[:, :, -padding:] = 0
        return mask


def unpad_image(tensor, original_size):
    """Unpads a PyTorch tensor of a padded and resized image.

    Args:
    tensor (torch.Tensor): The image tensor, assumed to be in CxHxW format.
    original_size (tuple): The original size of the image (height, width).

    Returns:
    torch.Tensor: The unpadded image tensor.
    """
    original_width, original_height = original_size
    current_height, current_width = tensor.shape[1:3]

    original_aspect_ratio = original_width / original_height
    current_aspect_ratio = current_width / current_height

    if original_aspect_ratio > current_aspect_ratio:
        scale_factor = current_width / original_width
        new_height = int(original_height * scale_factor)
        padding = (current_height - new_height) // 2
        unpadded_tensor = tensor[:, padding : current_height - padding, :]
    else:
        scale_factor = current_height / original_height
        new_width = int(original_width * scale_factor)
        padding = (current_width - new_width) // 2
        unpadded_tensor = tensor[:, :, padding : current_width - padding]

    return unpadded_tensor


class CambrianMetaForCausalLM(ABC):
    @abstractmethod
    def get_model(self):
        pass

    # def get_vision_tower(self):
    #     return self.get_model().get_vision_tower()

    def get_vision_tower_aux_list(self):
        return self.get_model().get_vision_tower_aux_list()

    def rearrange_vision_tower_features_train(
        self,
        vision_tower_aux_feature_list,
        vision_tower_aux_attention_masks_list,
        query_side_len,
    ):
        vision_tower_aux_feature_rearranged_list = []
        vision_tower_aux_attention_masks_rearranged_list = []
        bs = vision_tower_aux_feature_list[0].shape[0]
        for vision_tower_aux_feature, vision_tower_aux_attention_masks in zip(
            vision_tower_aux_feature_list, vision_tower_aux_attention_masks_list
        ):
            aux_height = aux_width = int(vision_tower_aux_feature.shape[1] ** 0.5)
            assert (aux_height // query_side_len) * query_side_len == aux_height

            reduce_factor = aux_height // query_side_len
            vision_tower_aux_feature_rearranged = vision_tower_aux_feature.view(
                bs, query_side_len, reduce_factor, query_side_len, reduce_factor, -1
            )
            vision_tower_aux_feature_rearranged = (
                vision_tower_aux_feature_rearranged.permute(0, 1, 3, 2, 4, 5)
                .contiguous()
                .flatten(0, 2)
                .flatten(1, 2)
            )

            vision_tower_aux_attention_masks_rearranged = (
                vision_tower_aux_attention_masks.view(
                    bs * query_side_len * query_side_len, reduce_factor * reduce_factor
                )
            )

            vision_tower_aux_feature_rearranged_list.append(
                vision_tower_aux_feature_rearranged
            )
            vision_tower_aux_attention_masks_rearranged_list.append(
                vision_tower_aux_attention_masks_rearranged
            )
        return (
            vision_tower_aux_feature_rearranged_list,
            vision_tower_aux_attention_masks_rearranged_list,
        )

    def rearrange_vision_tower_features_inference(
        self, vision_tower_aux_feature_list, query_side_len, image_sizes, unpad=False
    ):
        vision_tower_aux_feature_rearranged_list = []
        vision_tower_aux_attention_masks_rearranged_list = []
        bs = vision_tower_aux_feature_list[0].shape[0]
        for vision_tower_aux_feature in vision_tower_aux_feature_list:
            aux_height = aux_width = int(vision_tower_aux_feature.shape[1] ** 0.5)
            assert (aux_height // query_side_len) * query_side_len == aux_height

            reduce_factor = aux_height // query_side_len

            vision_tower_aux_feature_rearranged = []
            vision_tower_aux_attention_masks_rearranged = []
            for batch_i in range(bs):
                image_size = image_sizes[batch_i]
                cur_vision_tower_aux_feature = vision_tower_aux_feature[batch_i]

                cur_vision_tower_aux_attention_masks_rearranged = torch.ones(
                    (1, aux_height, aux_width),
                    dtype=torch.bool,
                    device=cur_vision_tower_aux_feature.device,
                )
                cur_vision_tower_aux_feature_rearranged = (
                    cur_vision_tower_aux_feature.view(
                        1,
                        query_side_len,
                        reduce_factor,
                        query_side_len,
                        reduce_factor,
                        -1,
                    )
                )
                cur_vision_tower_aux_feature_rearranged = (
                    cur_vision_tower_aux_feature_rearranged.permute(
                        0, 1, 3, 2, 4, 5
                    ).contiguous()
                )
                if unpad:
                    cur_vision_tower_aux_feature_rearranged = unpad_image(
                        cur_vision_tower_aux_feature_rearranged, image_size
                    )
                cur_vision_tower_aux_feature_rearranged = (
                    cur_vision_tower_aux_feature_rearranged.flatten(0, 2).flatten(1, 2)
                )  # query_side_len*query_side_len X reduce_factor*reduce_factor X C

                cur_vision_tower_aux_attention_masks_rearranged = unmask_attention_mask(
                    cur_vision_tower_aux_attention_masks_rearranged, image_size
                )
                cur_vision_tower_aux_attention_masks_rearranged = (
                    cur_vision_tower_aux_attention_masks_rearranged.view(
                        1, query_side_len, reduce_factor, query_side_len, reduce_factor
                    )
                    .permute(0, 1, 3, 2, 4)
                    .contiguous()
                )
                if unpad:
                    cur_vision_tower_aux_attention_masks_rearranged = unpad_image(
                        cur_vision_tower_aux_attention_masks_rearranged, image_size
                    )
                cur_vision_tower_aux_attention_masks_rearranged = (
                    cur_vision_tower_aux_attention_masks_rearranged.flatten(
                        0, 2
                    ).flatten(1, 2)
                )

                cur_vision_tower_aux_attention_masks_rearranged[
                    cur_vision_tower_aux_attention_masks_rearranged.sum(-1) == 0
                ] = True

                vision_tower_aux_feature_rearranged.append(
                    cur_vision_tower_aux_feature_rearranged
                )
                vision_tower_aux_attention_masks_rearranged.append(
                    cur_vision_tower_aux_attention_masks_rearranged
                )

            vision_tower_aux_feature_rearranged = torch.cat(
                vision_tower_aux_feature_rearranged, 0
            )
            vision_tower_aux_attention_masks_rearranged = torch.cat(
                vision_tower_aux_attention_masks_rearranged, 0
            )

            vision_tower_aux_feature_rearranged_list.append(
                vision_tower_aux_feature_rearranged
            )
            vision_tower_aux_attention_masks_rearranged_list.append(
                vision_tower_aux_attention_masks_rearranged
            )

        return (
            vision_tower_aux_feature_rearranged_list,
            vision_tower_aux_attention_masks_rearranged_list,
        )

    def encode_images(self, image_aux_list):
        vision_tower_aux_list = self.get_model().get_vision_tower_aux_list()
        image_aux_features_list = []
        for image_aux, vision_tower_aux in zip(image_aux_list, vision_tower_aux_list):
            image_aux_features = vision_tower_aux(image_aux)
            image_aux_features_list.append(image_aux_features)
        return image_aux_features_list

    def prepare_inputs_labels_for_multimodal(
        self,
        input_ids,
        position_ids,
        attention_mask,
        past_key_values,
        labels,
        images,
        image_aux_attention_masks_list=None,
        image_sizes=None,
    ):
        # vision_tower = self.get_vision_tower()
        vision_tower_aux_list = self.get_model().get_vision_tower_aux_list()
        if vision_tower_aux_list is None or images is None or input_ids.shape[1] == 1:
            return (
                input_ids,
                position_ids,
                attention_mask,
                past_key_values,
                None,
                labels,
                None,
                None,
                None,
                None,
            )

        image_aux_list = images

        bs = image_aux_list[0].shape[0]
        dtype = image_aux_list[0].dtype

        image_token_len = self.get_model().config.image_token_len
        query_num_list = self.get_model().config.query_num_list

        final_height = final_width = int(image_token_len**0.5)

        final_image_features_list = []

        # only needed for sva
        vision_tower_aux_feature_list_final = None
        vision_tower_aux_attention_masks_list_final = None
        global_context_feature_final = None

        image_aux_features_list = self.encode_images(image_aux_list)

        if self.get_model().config.mm_projector_type == "sva":
            vision_tower_aux_feature_list = []
            vision_tower_aux_attention_masks_list = []
            # get vision tokens from each vision tower
            for aux_i in range(len(vision_tower_aux_list)):
                image_aux_features = image_aux_features_list[aux_i]

                image_aux_features = getattr(
                    self.get_model(), f"mm_projector_aux_{aux_i}"
                )(image_aux_features).to(dtype)
                if aux_i == 0:
                    global_context_feature = image_aux_features.mean(1).view(
                        bs, 1, 1, -1
                    )

                vision_tower_aux_feature_list.append(image_aux_features)

            # perform vision sampling for each query group
            for query_group_i, query_num in enumerate(query_num_list):
                query_features_i = (
                    self.get_model()
                    .vision_query[query_group_i, :]
                    .view(1, 1, 1, -1)
                    .expand(bs, query_num, -1, -1)
                )
                global_context_feature_i = global_context_feature.expand(
                    -1, query_num, 1, -1
                ).flatten(0, 1)
                query_side_len = int(query_num**0.5)
                if IS_XLA_AVAILABLE:
                    (
                        vision_tower_aux_feature_list_i,
                        vision_tower_aux_attention_masks_list_i,
                    ) = self.rearrange_vision_tower_features_train(
                        vision_tower_aux_feature_list,
                        image_aux_attention_masks_list,
                        query_side_len,
                    )
                else:
                    (
                        vision_tower_aux_feature_list_i,
                        vision_tower_aux_attention_masks_list_i,
                    ) = self.rearrange_vision_tower_features_inference(
                        vision_tower_aux_feature_list, query_side_len, image_sizes
                    )

                query_features_i = getattr(
                    self.get_model(), f"vision_sampler_{query_group_i}"
                )(
                    query_features_i.flatten(0, 1),
                    global_context_feature_i,
                    *vision_tower_aux_feature_list_i,
                    *vision_tower_aux_attention_masks_list_i,
                )
                query_features_i = query_features_i.view(bs, query_num, -1)
                # interpolate to the final target size
                if query_side_len != final_height:
                    query_features_i = (
                        query_features_i.permute(0, 2, 1)
                        .contiguous()
                        .view(bs, -1, query_side_len, query_side_len)
                    )
                    query_features_i = F.interpolate(
                        query_features_i.float(),
                        size=(final_height, final_width),
                        mode="bilinear",
                        align_corners=False,
                    ).to(dtype=query_features_i.dtype)
                    query_features_i = (
                        query_features_i.permute(0, 2, 3, 1).contiguous().flatten(1, 2)
                    )
                final_image_features_list.append(query_features_i)

            if IS_XLA_AVAILABLE:
                (
                    vision_tower_aux_feature_list_final,
                    vision_tower_aux_attention_masks_list_final,
                ) = self.rearrange_vision_tower_features_train(
                    vision_tower_aux_feature_list,
                    image_aux_attention_masks_list,
                    final_height,
                )
                global_context_feature_final = global_context_feature.expand(
                    -1, final_height * final_width, 1, -1
                ).flatten(0, 1)
        else:
            final_image_features_list = image_aux_features_list

        image_features = torch.cat(final_image_features_list, -1)
        image_features = self.get_model().mm_projector(image_features).to(dtype)

        if IS_XLA_AVAILABLE:
            image_features = image_features.view(
                image_features.shape[0], final_height, final_width, -1
            )
            image_features = torch.cat(
                (
                    image_features,
                    self.model.image_newline[None, None, None, :].expand(
                        image_features.shape[0], final_height, 1, -1
                    ),
                ),
                dim=2,
            )
            image_features = image_features.flatten(1, 2)
            final_size = [(final_height, final_width)] * bs

        else:
            image_features = image_features.view(bs, final_height, final_width, -1)
            image_features_unpadded = []
            final_size = []
            if self.get_model().config.mm_projector_type == "sva":
                (
                    vision_tower_aux_feature_list_final,
                    vision_tower_aux_attention_masks_list_final,
                ) = self.rearrange_vision_tower_features_inference(
                    vision_tower_aux_feature_list, final_height, image_sizes, unpad=True
                )
                global_context_feature_final = []
            for batch_i in range(bs):
                cur_image_feature = image_features[batch_i]
                image_size = image_sizes[batch_i]

                cur_image_feature = unpad_image(
                    cur_image_feature.unsqueeze(0), image_size
                )

                cur_h, cur_w = cur_image_feature.shape[1:3]
                final_size.append((cur_h, cur_w))
                cur_image_feature = cur_image_feature.view(1, cur_h, cur_w, -1)
                cur_image_feature = torch.cat(
                    (
                        cur_image_feature,
                        self.model.image_newline.view(1, 1, 1, -1)
                        .expand(1, cur_h, 1, -1)
                        .to(cur_image_feature.device),
                    ),
                    dim=2,
                )
                cur_image_feature = cur_image_feature.flatten(1, 2)
                image_features_unpadded.append(cur_image_feature.squeeze(0))

                if self.get_model().config.mm_projector_type == "sva":
                    cur_global_context_feature = global_context_feature[batch_i].expand(
                        cur_h * cur_w, 1, -1
                    )
                    global_context_feature_final.append(cur_global_context_feature)
            if self.get_model().config.mm_projector_type == "sva":
                global_context_feature_final = torch.cat(
                    global_context_feature_final, 0
                )

            image_features = image_features_unpadded

        # TODO: image start / end is not implemented here to support pretraining.
        if getattr(self.config, "tune_mm_mlp_adapter", False) and getattr(
            self.config, "mm_use_im_start_end", False
        ):
            raise NotImplementedError

        if IS_XLA_AVAILABLE:
            # embed the input_ids
            new_input_ids_padded_for_emb = torch.where(
                input_ids == IMAGE_TOKEN_INDEX, 0, input_ids
            )
            input_embeds = self.get_model().embed_tokens(new_input_ids_padded_for_emb)
            new_input_embeds = []
            cur_image_idx = 0
            # insert the image embeddings
            for batch_idx, (cur_input_embeds, cur_input_ids) in enumerate(
                zip(input_embeds, input_ids)
            ):
                num_images = (cur_input_ids == IMAGE_TOKEN_INDEX).sum()
                if num_images == 0:
                    cur_image_idx += 1
                    new_input_embeds.append(cur_input_embeds)
                    continue

                image_token_indices = (
                    [-1]
                    + torch.where(cur_input_ids == IMAGE_TOKEN_INDEX)[0].tolist()
                    + [cur_input_ids.shape[0]]
                )

                cur_input_embeds_im_replaced = []

                prev_image_length = 0
                for i in range(len(image_token_indices) - 1):
                    # skip the image tokens (1 indicator + (image_length-1) paddings)
                    cur_input_embeds_im_replaced.append(
                        cur_input_embeds[
                            image_token_indices[i]
                            + 1
                            + prev_image_length : image_token_indices[i + 1]
                        ]
                    )
                    if i < len(image_token_indices) - 2:
                        cur_image_features = image_features[cur_image_idx]
                        prev_image_length = len(cur_image_features) - 1
                        cur_image_idx += 1
                        cur_input_embeds_im_replaced.append(cur_image_features)

                cur_input_embeds_im_replaced = [
                    x.to(self.device) for x in cur_input_embeds_im_replaced
                ]
                new_input_embeds.append(torch.cat(cur_input_embeds_im_replaced))

            new_input_embeds = torch.stack(new_input_embeds)
            return (
                None,
                position_ids,
                attention_mask,
                past_key_values,
                new_input_embeds,
                labels,
                vision_tower_aux_feature_list_final,
                vision_tower_aux_attention_masks_list_final,
                final_size,
                global_context_feature_final,
            )

        else:
            # Let's just add dummy tensors if they do not exist,
            # it is a headache to deal with None all the time.
            # But it is not ideal, and if you have a better idea,
            # please open an issue / submit a PR, thanks.
            _labels = labels
            _position_ids = position_ids
            _attention_mask = attention_mask
            if attention_mask is None:
                attention_mask = torch.ones_like(input_ids, dtype=torch.bool)
            else:
                attention_mask = attention_mask.bool()
            if position_ids is None:
                position_ids = torch.arange(
                    0, input_ids.shape[1], dtype=torch.long, device=input_ids.device
                )
            if labels is None:
                labels = torch.full_like(input_ids, IGNORE_INDEX)

            # remove the padding using attention_mask -- FIXME
            _input_ids = input_ids
            input_ids = [
                cur_input_ids[cur_attention_mask]
                for cur_input_ids, cur_attention_mask in zip(input_ids, attention_mask)
            ]
            labels = [
                cur_labels[cur_attention_mask]
                for cur_labels, cur_attention_mask in zip(labels, attention_mask)
            ]

            new_input_embeds = []
            new_labels = []
            cur_image_idx = 0
            for batch_idx, cur_input_ids in enumerate(input_ids):
                num_images = (cur_input_ids == IMAGE_TOKEN_INDEX).sum()
                if num_images == 0:
                    cur_image_features = image_features[cur_image_idx]
                    cur_input_embeds_1 = self.get_model().embed_tokens(cur_input_ids)
                    cur_input_embeds = torch.cat(
                        [cur_input_embeds_1, cur_image_features[0:0]], dim=0
                    )
                    new_input_embeds.append(cur_input_embeds)
                    new_labels.append(labels[batch_idx])
                    cur_image_idx += 1
                    continue

                image_token_indices = (
                    [-1]
                    + torch.where(cur_input_ids == IMAGE_TOKEN_INDEX)[0].tolist()
                    + [cur_input_ids.shape[0]]
                )
                cur_input_ids_noim = []
                cur_labels = labels[batch_idx]
                cur_labels_noim = []
                for i in range(len(image_token_indices) - 1):
                    cur_input_ids_noim.append(
                        cur_input_ids[
                            image_token_indices[i] + 1 : image_token_indices[i + 1]
                        ]
                    )
                    cur_labels_noim.append(
                        cur_labels[
                            image_token_indices[i] + 1 : image_token_indices[i + 1]
                        ]
                    )
                split_sizes = [x.shape[0] for x in cur_labels_noim]
                cur_input_embeds = self.get_model().embed_tokens(
                    torch.cat(cur_input_ids_noim)
                )
                cur_input_embeds_no_im = torch.split(
                    cur_input_embeds, split_sizes, dim=0
                )
                cur_new_input_embeds = []
                cur_new_labels = []

                for i in range(num_images + 1):
                    cur_new_input_embeds.append(cur_input_embeds_no_im[i])
                    cur_new_labels.append(cur_labels_noim[i])
                    if i < num_images:
                        cur_image_features = image_features[cur_image_idx]
                        cur_image_idx += 1
                        cur_new_input_embeds.append(cur_image_features)
                        cur_new_labels.append(
                            torch.full(
                                (cur_image_features.shape[0],),
                                IGNORE_INDEX,
                                device=cur_labels.device,
                                dtype=cur_labels.dtype,
                            )
                        )

                cur_new_input_embeds = [x.to(self.device) for x in cur_new_input_embeds]

                cur_new_input_embeds = torch.cat(cur_new_input_embeds)
                cur_new_labels = torch.cat(cur_new_labels)

                new_input_embeds.append(cur_new_input_embeds)
                new_labels.append(cur_new_labels)

            # Truncate sequences to max length as image embeddings can make the sequence longer
            tokenizer_model_max_length = getattr(
                self.config, "tokenizer_model_max_length", None
            )
            if tokenizer_model_max_length is not None:
                new_input_embeds = [
                    x[:tokenizer_model_max_length] for x in new_input_embeds
                ]
                new_labels = [x[:tokenizer_model_max_length] for x in new_labels]

            # Combine them
            max_len = max(x.shape[0] for x in new_input_embeds)
            batch_size = len(new_input_embeds)

            new_input_embeds_padded = []
            new_labels_padded = torch.full(
                (batch_size, max_len),
                IGNORE_INDEX,
                dtype=new_labels[0].dtype,
                device=new_labels[0].device,
            )
            attention_mask = torch.zeros(
                (batch_size, max_len),
                dtype=attention_mask.dtype,
                device=attention_mask.device,
            )
            position_ids = torch.zeros(
                (batch_size, max_len),
                dtype=position_ids.dtype,
                device=position_ids.device,
            )

            for i, (cur_new_embed, cur_new_labels) in enumerate(
                zip(new_input_embeds, new_labels)
            ):
                cur_len = cur_new_embed.shape[0]
                if getattr(self.config, "tokenizer_padding_side", "right") == "left":
                    new_input_embeds_padded.append(
                        torch.cat(
                            (
                                torch.zeros(
                                    (max_len - cur_len, cur_new_embed.shape[1]),
                                    dtype=cur_new_embed.dtype,
                                    device=cur_new_embed.device,
                                ),
                                cur_new_embed,
                            ),
                            dim=0,
                        )
                    )
                    if cur_len > 0:
                        new_labels_padded[i, -cur_len:] = cur_new_labels
                        attention_mask[i, -cur_len:] = True
                        position_ids[i, -cur_len:] = torch.arange(
                            0,
                            cur_len,
                            dtype=position_ids.dtype,
                            device=position_ids.device,
                        )
                else:
                    new_input_embeds_padded.append(
                        torch.cat(
                            (
                                cur_new_embed,
                                torch.zeros(
                                    (max_len - cur_len, cur_new_embed.shape[1]),
                                    dtype=cur_new_embed.dtype,
                                    device=cur_new_embed.device,
                                ),
                            ),
                            dim=0,
                        )
                    )
                    if cur_len > 0:
                        new_labels_padded[i, :cur_len] = cur_new_labels
                        attention_mask[i, :cur_len] = True
                        position_ids[i, :cur_len] = torch.arange(
                            0,
                            cur_len,
                            dtype=position_ids.dtype,
                            device=position_ids.device,
                        )

            new_input_embeds = torch.stack(new_input_embeds_padded, dim=0)

            if _labels is None:
                new_labels = None
            else:
                new_labels = new_labels_padded

            if _attention_mask is None:
                attention_mask = None
            else:
                attention_mask = attention_mask.to(dtype=_attention_mask.dtype)

            if _position_ids is None:
                position_ids = None

            return (
                None,
                position_ids,
                attention_mask,
                past_key_values,
                new_input_embeds,
                new_labels,
                vision_tower_aux_feature_list_final,
                vision_tower_aux_attention_masks_list_final,
                final_size,
                global_context_feature_final,
            )

    def initialize_vision_tokenizer(self, model_args, tokenizer):
        if model_args.mm_use_im_patch_token:
            tokenizer.add_tokens([DEFAULT_IMAGE_PATCH_TOKEN], special_tokens=True)
            self.resize_token_embeddings(len(tokenizer))

        if model_args.mm_use_im_start_end:
            num_new_tokens = tokenizer.add_tokens(
                [DEFAULT_IM_START_TOKEN, DEFAULT_IM_END_TOKEN], special_tokens=True
            )
            self.resize_token_embeddings(len(tokenizer))

            if num_new_tokens > 0:
                input_embeddings = self.get_input_embeddings().weight.data
                output_embeddings = self.get_output_embeddings().weight.data

                input_embeddings_avg = input_embeddings[:-num_new_tokens].mean(
                    dim=0, keepdim=True
                )
                output_embeddings_avg = output_embeddings[:-num_new_tokens].mean(
                    dim=0, keepdim=True
                )

                input_embeddings[-num_new_tokens:] = input_embeddings_avg
                output_embeddings[-num_new_tokens:] = output_embeddings_avg

            if model_args.tune_mm_mlp_adapter:
                for p in self.get_input_embeddings().parameters():
                    p.requires_grad = True
                for p in self.get_output_embeddings().parameters():
                    p.requires_grad = False

            if model_args.pretrain_mm_mlp_adapter:
                mm_projector_weights = torch.load(
                    model_args.pretrain_mm_mlp_adapter, map_location="cpu"
                )
                embed_tokens_weight = mm_projector_weights["model.embed_tokens.weight"]
                assert num_new_tokens == 2
                if input_embeddings.shape == embed_tokens_weight.shape:
                    input_embeddings[-num_new_tokens:] = embed_tokens_weight[
                        -num_new_tokens:
                    ]
                elif embed_tokens_weight.shape[0] == num_new_tokens:
                    input_embeddings[-num_new_tokens:] = embed_tokens_weight
                else:
                    raise ValueError(
                        f"Unexpected embed_tokens_weight shape. Pretrained: {embed_tokens_weight.shape}. Current: {input_embeddings.shape}. Numer of new tokens: {num_new_tokens}."
                    )
        elif model_args.mm_use_im_patch_token:
            if model_args.tune_mm_mlp_adapter:
                for p in self.get_input_embeddings().parameters():
                    p.requires_grad = False
                for p in self.get_output_embeddings().parameters():
                    p.requires_grad = False



import re
import tempfile
from pathlib import Path

import pytest

from oumi.core.analyze.dataset_analyzer import (
    ConversationAnalysisResult,
    MessageAnalysisResult,
)
from oumi.core.configs import EvaluationBackend, EvaluationConfig, EvaluationTaskParams
from oumi.core.evaluation.evaluation_result import EvaluationResult
from oumi.core.registry import (
    REGISTRY,
    Registry,
    RegistryType,
    register,
    register_dataset,
    register_evaluation_function,
    register_sample_analyzer,
)


@pytest.fixture(autouse=True)
def mock_env_vars(monkeypatch):
    monkeypatch.setenv("OUMI_EXTRA_DEPS_FILE", "")


@pytest.fixture(autouse=True)
def cleanup():
    snapshot = Registry()
    for reg_type in RegistryType:
        for key, value in REGISTRY.get_all(reg_type).items():
            snapshot.register(key, reg_type, value)
    # Clear the registry before each test.
    REGISTRY.clear()
    REGISTRY._initialized = False
    yield
    # Clear the registry after each test.
    REGISTRY.clear()
    REGISTRY._initialized = False
    # Restore the registry after each test.
    for reg_type in RegistryType:
        for key, value in snapshot.get_all(reg_type).items():
            REGISTRY.register(key, reg_type, value)


def test_registry_cloud_builder():
    class DummyClass:
        pass

    @register("dummy_class", RegistryType.CLOUD)
    def dummy_builder():
        return DummyClass()

    assert REGISTRY.contains("dummy_class", RegistryType.CLOUD)
    assert REGISTRY.get("dummy_class", RegistryType.CLOUD) == dummy_builder
    assert not REGISTRY.contains("some_other_class", RegistryType.CLOUD)


def test_registry_cloud_builder_get_all():
    class DummyClass:
        pass

    @register("dummy_class", RegistryType.CLOUD)
    def dummy_builder():
        return DummyClass()

    class SomeClass:
        pass

    @register("another_one", RegistryType.CLOUD)
    def another_builder():
        return SomeClass()

    class LastClass:
        pass

    @register("finally", RegistryType.CLOUD)
    def last_builder():
        return LastClass()

    all_builders = REGISTRY.get_all(RegistryType.CLOUD).values()
    assert list(all_builders) == [dummy_builder, another_builder, last_builder]


def test_registry_model_class():
    @register("dummy_class", RegistryType.MODEL)
    class DummyClass:
        pass

    assert REGISTRY.contains("dummy_class", RegistryType.MODEL)
    assert not REGISTRY.contains("dummy_class", RegistryType.MODEL_CONFIG)
    assert REGISTRY.get("dummy_class", RegistryType.MODEL) == DummyClass


def test_registry_model_config_class():
    @register("dummy_config_class", RegistryType.MODEL_CONFIG)
    class DummyConfigClass:
        pass

    assert REGISTRY.contains("dummy_config_class", RegistryType.MODEL_CONFIG)
    assert not REGISTRY.contains("dummy_config_class", RegistryType.MODEL)
    assert (
        REGISTRY.get("dummy_config_class", RegistryType.MODEL_CONFIG)
        == DummyConfigClass
    )


def test_registry_failure_register_class_twice():
    @register("another_dummy_class", RegistryType.MODEL)
    class DummyClass:
        pass

    with pytest.raises(ValueError) as exception_info:

        @register("another_dummy_class", RegistryType.MODEL)
        class AnotherDummyClass:
            pass

    assert "already registered" in str(exception_info.value)


def test_registry_failure_get_unregistered_class():
    assert not REGISTRY.contains("unregistered_class", RegistryType.MODEL)
    assert not REGISTRY.get(name="unregistered_class", type=RegistryType.MODEL)

    with pytest.raises(KeyError) as exception_info:
        REGISTRY["unregistered_class", RegistryType.MODEL]

    assert "does not exist" in str(exception_info.value)


def test_registry_model():
    @register("oumi/dummy", RegistryType.MODEL_CONFIG)
    class DummyModelConfig:
        pass

    @register("oumi/dummy", RegistryType.MODEL)
    class DummyModelClass:
        pass

    model_class = REGISTRY.get_model("oumi/dummy")
    assert model_class
    assert model_class == DummyModelClass

    model_config = REGISTRY.get_model_config("oumi/dummy")
    assert model_config
    assert model_config == DummyModelConfig


def test_registry_failure_model_not_present_in_registry():
    @register("oumi/dummy1", RegistryType.MODEL_CONFIG)
    class DummyModelConfig:
        pass

    @register("oumi/dummy2", RegistryType.MODEL)
    class DummyModelClass:
        pass

    # Non-existent model (without exception).
    assert REGISTRY.get_model(name="oumi/yet_another_dummy") is None

    # Non-existent model (with exception).
    with pytest.raises(KeyError) as exception_info:
        REGISTRY["oumi/yet_another_dummy", RegistryType.MODEL]

    assert "does not exist" in str(exception_info.value)

    # Incomplete model (without exception).
    assert REGISTRY.get_model(name="oumi/dummy1") is None
    assert REGISTRY.get_model_config(name="oumi/dummy2") is None


@pytest.mark.parametrize(
    "registry_type", [RegistryType.METRICS_FUNCTION, RegistryType.REWARD_FUNCTION]
)
def test_registry_function(registry_type: RegistryType):
    @register("dummy_fn", registry_type)
    def dummy_function():
        pass

    @register("number2", registry_type)
    def dummy_function2():
        pass

    assert REGISTRY.contains("dummy_fn", registry_type)
    assert REGISTRY.get("dummy_fn", registry_type) == dummy_function

    assert REGISTRY.contains("number2", registry_type)
    assert REGISTRY.get("number2", registry_type) == dummy_function2


@pytest.mark.parametrize(
    "registry_type", [RegistryType.METRICS_FUNCTION, RegistryType.REWARD_FUNCTION]
)
def test_registry_metrics_function_non_callable(registry_type: RegistryType):
    class FooNonCallable:
        pass

    with pytest.raises(
        ValueError,
        match=re.escape(
            f"Registry: `foo_non_callable` of `{registry_type}` must be callable"
        ),
    ):
        REGISTRY.register("foo_non_callable", registry_type, FooNonCallable())

    with pytest.raises(
        ValueError,
        match=re.escape(
            f"Registry: `integer_func` of `{registry_type}` must be callable"
        ),
    ):
        REGISTRY.register("integer_func", registry_type, 99)

    with pytest.raises(
        ValueError,
        match=re.escape(f"Registry: `none_func` of `{registry_type}` must be callable"),
    ):
        REGISTRY.register("none_func", registry_type, None)


def test_registry_failure_metrics_function_not_present():
    @register("dummy_fn", RegistryType.METRICS_FUNCTION)
    def dummy_function():
        pass

    @register("number2", RegistryType.METRICS_FUNCTION)
    def dummy_function2():
        pass

    # Non-existent function (without exception).
    assert REGISTRY.get_model(name="dummy") is None

    # Non-existent function (with exception).
    with pytest.raises(KeyError) as exception_info:
        REGISTRY["yet_another_dummy", RegistryType.METRICS_FUNCTION]

    assert "does not exist" in str(exception_info.value)

    # Incomplete function name (without exception).
    assert REGISTRY.get_model(name="dummy_") is None
    assert REGISTRY.get_model_config(name="number") is None


def test_registry_datasets():
    dataset_name = "test_dataset"
    subset_name = "test_subset"

    @register_dataset(dataset_name)
    class TestDataset:
        pass

    @register_dataset(dataset_name, subset=subset_name)
    class TestDatasetSubset:
        pass

    # Dataset class should be registered.
    # If no subset is specified, the main dataset class should be returned.
    dataset_class = REGISTRY.get_dataset(dataset_name)
    assert dataset_class is not None
    assert dataset_class == TestDataset

    # Getting a dataset class for a registered subset
    # If a subset is specificed, we should get the subset class if one was registered
    dataset_subset_class = REGISTRY.get_dataset(dataset_name, subset=subset_name)
    assert dataset_subset_class is not None
    assert dataset_subset_class == TestDatasetSubset

    # Getting a dataset class for a non-registered subset
    # In this case we should get the main class
    dataset_subset_class = REGISTRY.get_dataset(
        dataset_name, subset="non_existent_subset"
    )
    assert dataset_subset_class is not None
    # If a subset is specificed, we should get the subset class if one was registered
    assert dataset_subset_class == TestDataset


@pytest.mark.parametrize(
    "registry_type",
    [
        RegistryType.CLOUD,
        RegistryType.MODEL,
        RegistryType.MODEL_CONFIG,
        RegistryType.METRICS_FUNCTION,
        RegistryType.DATASET,
    ],
)
def test_registry_case_insensitive(registry_type):
    class DummyClass:
        pass

    @register("Dummy_Class", registry_type)
    def dummy_builder():
        return DummyClass()

    # Check if the registry contains the key, regardless of case
    assert REGISTRY.contains("dummy_class", registry_type)
    assert REGISTRY.contains("DUMMY_CLASS", registry_type)
    assert REGISTRY.contains("Dummy_Class", registry_type)

    # Check if we can retrieve the builder using different cases
    assert REGISTRY.get("dummy_class", registry_type) == dummy_builder
    assert REGISTRY.get("DUMMY_CLASS", registry_type) == dummy_builder
    assert REGISTRY.get("Dummy_Class", registry_type) == dummy_builder

    # Check if the registry does not contain the key, regardless of case
    assert not REGISTRY.contains("dummy_class_not_registered", registry_type)
    assert not REGISTRY.contains("DUMMY_CLASS_NOT_REGISTERED", registry_type)
    assert not REGISTRY.contains("Dummy_Class_Not_Registered", registry_type)


def test_registry_case_insensitive_multiple_registrations():
    @register("Test_Class", RegistryType.MODEL)
    class TestClass1:
        pass

    with pytest.raises(ValueError, match="already registered"):

        @register("test_class", RegistryType.MODEL)
        class TestClass2:
            pass


def test_registry_case_insensitive_get_all():
    @register("Class_One", RegistryType.CLOUD)
    def builder_one():
        pass

    @register("CLASS_TWO", RegistryType.CLOUD)
    def builder_two():
        pass

    all_builders = REGISTRY.get_all(RegistryType.CLOUD)
    assert set(all_builders.keys()) == {"class_one", "class_two"}
    assert list(all_builders.values()) == [builder_one, builder_two]


def test_registry_contains_initialization():
    assert REGISTRY._initialized is False
    assert len(REGISTRY._registry) == 0
    _ = REGISTRY.contains("foo", RegistryType.CLOUD)
    assert REGISTRY._initialized


def test_registry_register_initialization():
    assert REGISTRY._initialized is False
    assert len(REGISTRY._registry) == 0
    _ = REGISTRY.register("foo", RegistryType.CLOUD, "bar")
    assert REGISTRY._initialized


def test_registry_get_initialization():
    assert REGISTRY._initialized is False
    assert len(REGISTRY._registry) == 0
    _ = REGISTRY.get("foo", RegistryType.CLOUD)
    assert REGISTRY._initialized


def test_registry_get_all_initialization():
    assert REGISTRY._initialized is False
    assert len(REGISTRY._registry) == 0
    _ = REGISTRY.get_all(RegistryType.CLOUD)
    assert REGISTRY._initialized


def test_registry_user_classes(monkeypatch):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        req_file = Path(output_temp_dir) / "requirements.txt"
        file_1 = Path(output_temp_dir) / "file_1.py"
        file_2 = Path(output_temp_dir) / "another_file.py"
        file_3 = Path(output_temp_dir) / "last_one.py"
        with monkeypatch.context() as mp:
            mp.setenv("OUMI_EXTRA_DEPS_FILE", str(req_file))
            with open(req_file, "w") as f:
                f.write(str(file_1) + "\n\n")  # Add an empty line
                f.write(str(file_2) + "\n")
                f.write(str(file_3) + "\n")
            with open(file_1, "w") as f:
                f.writelines(
                    [
                        "from oumi.core.registry import register, RegistryType\n",
                        "@register('file_1', RegistryType.CLOUD)\n",
                        "class FileOne:\n",
                        "    pass\n",
                    ]
                )
            with open(file_2, "w") as f:
                f.writelines(
                    [
                        "from oumi.core.registry import register, RegistryType\n",
                        "@register('file_2', RegistryType.MODEL)\n",
                        "class FileTwo:\n",
                        "    pass\n",
                    ]
                )
            with open(file_3, "w") as f:
                f.writelines(
                    [
                        "from oumi.core.registry import register, RegistryType\n",
                        "@register('file_3', RegistryType.METRICS_FUNCTION)\n",
                        "class FileThree:\n",
                        "    pass\n",
                    ]
                )
            assert not REGISTRY._initialized
            assert REGISTRY.contains("file_1", RegistryType.CLOUD)
            assert REGISTRY.contains("file_2", RegistryType.MODEL)
            assert REGISTRY.contains("file_3", RegistryType.METRICS_FUNCTION)
            assert REGISTRY._initialized


def test_registry_user_classes_empty_requirements(monkeypatch):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        req_file = Path(output_temp_dir) / "requirements.txt"
        with monkeypatch.context() as mp:
            mp.setenv("OUMI_EXTRA_DEPS_FILE", str(req_file))
            with open(req_file, "w") as f:
                f.write("\n")
            assert not REGISTRY._initialized
            assert not REGISTRY.contains("file_1", RegistryType.CLOUD)
            assert REGISTRY._initialized


def test_registry_user_classes_malformed_dep(monkeypatch):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        req_file = Path(output_temp_dir) / "requirements.txt"
        file_1 = Path(output_temp_dir) / "file_1.py"
        with monkeypatch.context() as mp:
            mp.setenv("OUMI_EXTRA_DEPS_FILE", str(req_file))
            with open(req_file, "w") as f:
                f.write(str(file_1) + "\n\n")  # Add an empty line
                f.write(str(Path(output_temp_dir) / "non_existent_file.py") + "\n")
            with open(file_1, "w") as f:
                f.writelines(
                    [
                        "fr om thisisbadpython import fakemodulethatfails\n",
                        "@register('file_1', RegistryType.CLOUD)\n",
                        "class FileOne:\n",
                        "    pass\n",
                    ]
                )
            assert not REGISTRY._initialized
            with pytest.raises(
                ImportError, match="Failed to load user-defined module:"
            ):
                REGISTRY.contains("file_1", RegistryType.CLOUD)


def test_registry_user_classes_missing_dep(monkeypatch):
    with tempfile.TemporaryDirectory() as output_temp_dir:
        req_file = Path(output_temp_dir) / "requirements.txt"
        file_1 = Path(output_temp_dir) / "file_1.py"
        with monkeypatch.context() as mp:
            mp.setenv("OUMI_EXTRA_DEPS_FILE", str(req_file))
            with open(req_file, "w") as f:
                f.write(str(file_1) + "\n\n")  # Add an empty line
                f.write(str(Path(output_temp_dir) / "non_existent_file.py") + "\n")
            with open(file_1, "w") as f:
                f.writelines(
                    [
                        "from oumi.core.registry import register, RegistryType\n",
                        "@register('file_1', RegistryType.CLOUD)\n",
                        "class FileOne:\n",
                        "    pass\n",
                    ]
                )
            assert not REGISTRY._initialized
            with pytest.raises(
                ImportError, match="Failed to load user-defined module:"
            ):
                REGISTRY.contains("file_1", RegistryType.CLOUD)


# Tests for registering evaluation functions.
def test_register_evaluation_fn_with_annotations_happy_path():
    @register_evaluation_function("test_evaluation_fn")
    def oumi_test_evaluation_fn(
        task_params: EvaluationTaskParams,
        config: EvaluationConfig,
        optional_param: str,
    ) -> EvaluationResult:
        """Dummy evaluation function for unit testing."""
        assert task_params.evaluation_backend == EvaluationBackend.CUSTOM.value
        assert task_params.task_name == "test_evaluation_fn"
        assert config.run_name == "run_name_for_test_evaluation_fn"
        assert optional_param == "optional_param_value"

        return EvaluationResult(
            task_name=task_params.task_name,
            task_result={"result": "dummy_result"},
            backend_config={"config": "dummy_config"},
        )

    evaluation_fn = REGISTRY.get_evaluation_function("test_evaluation_fn")
    assert evaluation_fn
    evaluation_result = evaluation_fn(
        task_params=EvaluationTaskParams(
            task_name="test_evaluation_fn",
            evaluation_backend=EvaluationBackend.CUSTOM.value,
        ),
        config=EvaluationConfig(run_name="run_name_for_test_evaluation_fn"),
        optional_param="optional_param_value",
    )
    assert evaluation_result.task_name == "test_evaluation_fn"
    assert evaluation_result.task_result == {"result": "dummy_result"}
    assert evaluation_result.backend_config == {"config": "dummy_config"}


def test_register_evaluation_fn_without_annotations_happy_path():
    @register_evaluation_function("test_evaluation_fn")
    def oumi_test_evaluation_fn(task_params, config, optional_param):
        """Dummy evaluation function for unit testing."""
        assert task_params.evaluation_backend == EvaluationBackend.CUSTOM.value
        assert task_params.task_name == "test_evaluation_fn"
        assert config.run_name == "run_name_for_test_evaluation_fn"
        assert optional_param == "optional_param_value"

        return EvaluationResult(
            task_name=task_params.task_name,
            task_result={"result": "dummy_result"},
            backend_config={"config": "dummy_config"},
        )

    evaluation_fn = REGISTRY.get_evaluation_function("test_evaluation_fn")
    assert evaluation_fn
    evaluation_result = evaluation_fn(
        task_params=EvaluationTaskParams(
            task_name="test_evaluation_fn",
            evaluation_backend=EvaluationBackend.CUSTOM.value,
        ),
        config=EvaluationConfig(run_name="run_name_for_test_evaluation_fn"),
        optional_param="optional_param_value",
    )
    assert evaluation_result.task_name == "test_evaluation_fn"
    assert evaluation_result.task_result == {"result": "dummy_result"}
    assert evaluation_result.backend_config == {"config": "dummy_config"}


def test_register_evaluation_fn_without_inputs_happy_path():
    @register_evaluation_function("test_evaluation_fn")
    def oumi_test_evaluation_fn():
        """Dummy evaluation function for unit testing."""
        return EvaluationResult(
            task_name="unknown_task",
            task_result={"result": "dummy_result"},
            backend_config={"config": "dummy_config"},
        )

    evaluation_fn = REGISTRY.get_evaluation_function("test_evaluation_fn")
    assert evaluation_fn
    evaluation_result = evaluation_fn()
    assert evaluation_result.task_name == "unknown_task"
    assert evaluation_result.task_result == {"result": "dummy_result"}
    assert evaluation_result.backend_config == {"config": "dummy_config"}


# Tests for registering sample analyzers.
def test_registry_sample_analyzer():
    @register_sample_analyzer("dummy_analyzer")
    class DummyAnalyzer:
        def analyze_sample(
            self, conversation, tokenizer=None
        ) -> tuple[list[MessageAnalysisResult], ConversationAnalysisResult]:
            return (
                [
                    MessageAnalysisResult(
                        message_index=0,
                        role="user",
                        message_id="dummy_msg",
                        text_content="dummy_message",
                        analyzer_metrics={},
                    )
                ],
                ConversationAnalysisResult(analyzer_metrics={}),
            )

    assert REGISTRY.contains("dummy_analyzer", RegistryType.SAMPLE_ANALYZER)
    assert REGISTRY.get("dummy_analyzer", RegistryType.SAMPLE_ANALYZER) == DummyAnalyzer
    assert not REGISTRY.contains("some_other_analyzer", RegistryType.SAMPLE_ANALYZER)


def test_registry_sample_analyzer_get_all():
    @register_sample_analyzer("analyzer_one")
    class AnalyzerOne:
        def analyze_sample(
            self, conversation, tokenizer=None
        ) -> tuple[list[MessageAnalysisResult], ConversationAnalysisResult]:
            return (
                [
                    MessageAnalysisResult(
                        message_index=0,
                        role="user",
                        message_id="dummy_msg",
                        text_content="dummy_message",
                        analyzer_metrics={},
                    )
                ],
                ConversationAnalysisResult(analyzer_metrics={}),
            )

    @register_sample_analyzer("analyzer_two")
    class AnalyzerTwo:
        def analyze_sample(
            self, conversation, tokenizer=None
        ) -> tuple[list[MessageAnalysisResult], ConversationAnalysisResult]:
            return (
                [
                    MessageAnalysisResult(
                        message_index=0,
                        role="user",
                        message_id="dummy_msg",
                        text_content="dummy_message",
                        analyzer_metrics={},
                    )
                ],
                ConversationAnalysisResult(analyzer_metrics={}),
            )

    all_analyzers = REGISTRY.get_all(RegistryType.SAMPLE_ANALYZER).values()
    assert list(all_analyzers) == [AnalyzerOne, AnalyzerTwo]


def test_registry_sample_analyzer_failure_register_twice():
    @register_sample_analyzer("duplicate_analyzer")
    class DummyAnalyzer:
        def analyze_sample(
            self, conversation, tokenizer=None
        ) -> tuple[list[MessageAnalysisResult], ConversationAnalysisResult]:
            return (
                [
                    MessageAnalysisResult(
                        message_index=0,
                        role="user",
                        message_id="dummy_msg",
                        text_content="dummy_message",
                        analyzer_metrics={},
                    )
                ],
                ConversationAnalysisResult(analyzer_metrics={}),
            )

    with pytest.raises(ValueError):

        @register_sample_analyzer("duplicate_analyzer")
        class AnotherDummyAnalyzer:
            def analyze_sample(
                self, conversation, tokenizer=None
            ) -> tuple[list[MessageAnalysisResult], ConversationAnalysisResult]:
                return (
                    [
                        MessageAnalysisResult(
                            message_index=0,
                            role="user",
                            message_id="dummy_msg",
                            text_content="dummy_message",
                            analyzer_metrics={},
                        )
                    ],
                    ConversationAnalysisResult(analyzer_metrics={}),
                )
